[llvm] [LoopInterchange] Supported partially-perfect Loop Nests (PR #199511)

Rohit Garg via llvm-commits llvm-commits at lists.llvm.org
Sun Jul 19 23:34:18 PDT 2026


https://github.com/rohgarg-qual updated https://github.com/llvm/llvm-project/pull/199511

>From 876226ee4ed09180e3b1946a379603c34fe868af Mon Sep 17 00:00:00 2001
From: rohgarg <rohgarg at qti.qualcomm.com>
Date: Mon, 25 May 2026 03:09:29 -0700
Subject: [PATCH 1/9] [LoopInterchange] Supported Imperfect Loop Nests
MIME-Version: 1.0
Content-Type: text/plain; charset=UTF-8
Content-Transfer-Encoding: 8bit

In the current implementation when the outermost loop contains sibling sub-loops (i.e. multiple independent loop nests at the same depth), the pass silently skips the entire loop nest without attempting to interchange any of the valid sub-nests (See Below Example).

Example:
for(int i=0; i<n; i++){  // Loop 1
    for(int j=0; j<m; j++){  // Loop 2
         for(int r=0; r<m; r++){  // Loop 3
         // Do something
         }
    }
    for(int k=0; k<p; k++){  // Loop 4
        for(int l=0; l<p; l++){  // Loop 5
            //Access A[l][k]
        }
    }
}

In the Example , GCC 16 performs loop interchange on (Loop2, Loop3) or (Loop4, Loop5), provided the interchange is both legal and profitable.

Fixes: https://github.com/llvm/llvm-project/issues/196006
---
 .../lib/Transforms/Scalar/LoopInterchange.cpp | 104 ++++++++---
 .../LoopInterchange/imperfect-loop-nest.ll    | 176 ++++++++++++++++++
 .../LoopInterchange/large-nested-6d.ll        |   1 +
 3 files changed, 252 insertions(+), 29 deletions(-)
 create mode 100644 llvm/test/Transforms/LoopInterchange/imperfect-loop-nest.ll

diff --git a/llvm/lib/Transforms/Scalar/LoopInterchange.cpp b/llvm/lib/Transforms/Scalar/LoopInterchange.cpp
index 440ddb182c272..01de3ead6ad60 100644
--- a/llvm/lib/Transforms/Scalar/LoopInterchange.cpp
+++ b/llvm/lib/Transforms/Scalar/LoopInterchange.cpp
@@ -167,7 +167,8 @@ static bool inThisOrder(const Instruction *Src, const Instruction *Dst) {
 #endif
 
 static bool populateDependencyMatrix(CharMatrix &DepMatrix, unsigned Level,
-                                     Loop *L, DependenceInfo *DI,
+                                     Loop *L, ArrayRef<Loop *> LoopList,
+                                     LoopInfo *LI, DependenceInfo *DI,
                                      ScalarEvolution *SE,
                                      OptimizationRemarkEmitter *ORE) {
   using ValueVector = SmallVector<Value *, 16>;
@@ -175,19 +176,27 @@ static bool populateDependencyMatrix(CharMatrix &DepMatrix, unsigned Level,
   ValueVector MemInstr;
   unsigned NumInsts = 0;
 
-  // For each block.
-  for (BasicBlock *BB : L->blocks()) {
-    // Scan the BB and collect legal loads and stores.
-    for (Instruction &I : *BB) {
-      NumInsts++;
-      if (auto *Ld = dyn_cast<LoadInst>(&I)) {
-        if (!Ld->isSimple())
-          return false;
-        MemInstr.push_back(&I);
-      } else if (auto *St = dyn_cast<StoreInst>(&I)) {
-        if (!St->isSimple())
+  // Collect memory instructions from the BB which belongs to all loops in the LoopList
+  for (Loop *PathLoop : LoopList) {
+    for (BasicBlock *BB : PathLoop->getBlocksVector()) {
+      // In this iteration we need to handle the BB contained directly by Current Loop 
+      // and all other BB of subloops will be handled in its own in next iterations.
+      if (LI->getLoopFor(BB) != PathLoop)
+        continue;
+      // Scan the BB and collect legal loads and stores.
+      for (Instruction &I : *BB) {
+        if (!isa<Instruction>(I))
           return false;
-        MemInstr.push_back(&I);
+        NumInsts++;
+        if (auto *Ld = dyn_cast<LoadInst>(&I)) {
+          if (!Ld->isSimple())
+            return false;
+          MemInstr.push_back(&I);
+        } else if (auto *St = dyn_cast<StoreInst>(&I)) {
+          if (!St->isSimple())
+            return false;
+          MemInstr.push_back(&I);
+        }
       }
     }
   }
@@ -674,12 +683,59 @@ struct LoopInterchange {
     return processLoopList(LoopList);
   }
 
+  static SmallVector<SmallVector<Loop *, 8>, 4> collectRootToLeafPaths(Loop *Root) {
+    SmallVector<SmallVector<Loop *, 8>, 4> AllPaths;
+    // Stack stores {Loop, CurrentPath} pairs
+    SmallVector<std::pair<Loop *, SmallVector<Loop *, 8>>, 8> Stack;
+    Stack.push_back({Root, {Root}});
+    while (!Stack.empty()) {
+      auto [L, CurrentPath] = Stack.pop_back_val();
+      const auto &SubLoops = L->getSubLoops();
+      if (SubLoops.empty()) {
+        // Leaf node: save the path
+        AllPaths.push_back(CurrentPath);
+      } else {
+        for (Loop *Child : SubLoops) {
+          SmallVector<Loop *, 8> NewPath(CurrentPath);
+          NewPath.push_back(Child);
+          Stack.push_back({Child, NewPath});
+        }
+      }
+    }
+    return AllPaths;
+  }
+
   bool run(LoopNest &LN) {
-    SmallVector<Loop *, 8> LoopList(LN.getLoops());
-    for (unsigned I = 1; I < LoopList.size(); ++I)
-      if (LoopList[I]->getParentLoop() != LoopList[I - 1])
-        return false;
-    return processLoopList(LoopList);
+    auto Paths = collectRootToLeafPaths(&LN.getOutermostLoop());
+    // Consider below kernel
+    // for(int i=0; i<n; i++){  // Loop 1
+    //     for(int j=0; j<m; j++){  // Loop 2
+    //         for(int r=0; r<m; r++){  // Loop 3
+    //         // Do something
+    //         }
+    //     }
+    //     for(int k=0; k<p; k++){  // Loop 4
+    //         for(int l=0; l<p; l++){  // Loop 5
+    //             // Do something
+    //         }
+    //     }
+    // }
+    // Then Paths will contain:
+    // - [Loop1, Loop2, Loop3]
+    // - [Loop1, Loop4, Loop5]
+    bool Changed = false;
+    for (auto &Path : Paths) {
+      // Ensure minimum depth of the loop nest to do the interchange.
+      if (!hasSupportedLoopDepth(Path, *ORE))
+        continue;
+      // Ensure computable loop nest.
+      if (!isComputableLoopNest(&AR->SE, Path)) {
+        LLVM_DEBUG(dbgs() << "Not valid loop candidate for interchange\n");
+        continue;
+      }
+      Changed |= processLoopList(Path);
+    }
+    return Changed;
   }
 
   unsigned selectLoopForInterchange(ArrayRef<Loop *> LoopList) {
@@ -709,7 +765,7 @@ struct LoopInterchange {
     CharMatrix DependencyMatrix;
     Loop *OuterMostLoop = *(LoopList.begin());
     if (!populateDependencyMatrix(DependencyMatrix, LoopNestDepth,
-                                  OuterMostLoop, DI, SE, ORE)) {
+                                  OuterMostLoop, LoopList, LI, DI, SE, ORE)) {
       LLVM_DEBUG(dbgs() << "Populating dependency matrix failed\n");
       return false;
     }
@@ -2602,19 +2658,9 @@ PreservedAnalyses LoopInterchangePass::run(LoopNest &LN,
                                            LoopStandardAnalysisResults &AR,
                                            LPMUpdater &U) {
   Function &F = *LN.getParent();
-  SmallVector<Loop *, 8> LoopList(LN.getLoops());
 
   OptimizationRemarkEmitter ORE(&F);
 
-  // Ensure minimum depth of the loop nest to do the interchange.
-  if (!hasSupportedLoopDepth(LoopList, ORE))
-    return PreservedAnalyses::all();
-  // Ensure computable loop nest.
-  if (!isComputableLoopNest(&AR.SE, LoopList)) {
-    LLVM_DEBUG(dbgs() << "Not valid loop candidate for interchange\n");
-    return PreservedAnalyses::all();
-  }
-
   ORE.emit([&]() {
     return OptimizationRemarkAnalysis(DEBUG_TYPE, "Dependence",
                                       LN.getOutermostLoop().getStartLoc(),
diff --git a/llvm/test/Transforms/LoopInterchange/imperfect-loop-nest.ll b/llvm/test/Transforms/LoopInterchange/imperfect-loop-nest.ll
new file mode 100644
index 0000000000000..d552d32e46377
--- /dev/null
+++ b/llvm/test/Transforms/LoopInterchange/imperfect-loop-nest.ll
@@ -0,0 +1,176 @@
+; NOTE: Assertions have been autogenerated by utils/update_test_checks.py
+; RUN: opt -S -passes='loop-interchange' -cache-line-size=64 < %s | FileCheck %s
+
+;  // Loop nest with multiple loops at same depth
+;    for (i = 0; i < 64; i++)
+;    {
+;      for (j = 0; j < 64; j++)
+;        {
+;          #pragma clang loop unroll(disable)
+;          for (r = 0; r < 64; r++)
+;            {
+;              C[j][r] = B[j][r] + i;
+;            }
+;        }
+;
+;      for (k = 0; k < 64; k++)
+;        {
+;          #pragma clang loop unroll(disable)
+;          for (l = 0; l < 64; l++)
+;            {
+;              A[l][k] = A[l][k] + 1;
+;            }
+;        }
+;    }
+
+target datalayout = "e-m:e-p270:32:32-p271:32:32-p272:64:64-i8:8:32-i16:16:32-i64:64-i128:128-n32:64-S128-Fn32"
+target triple = "aarch64-unknown-linux-gnueabi"
+
+ at B = dso_local local_unnamed_addr global [64 x [64 x i32]] zeroinitializer, align 4
+ at C = dso_local local_unnamed_addr global [64 x [64 x i32]] zeroinitializer, align 4
+ at A = dso_local local_unnamed_addr global [64 x [64 x i32]] zeroinitializer, align 4
+
+; Function Attrs: nofree norecurse nosync nounwind memory(readwrite, argmem: none, inaccessiblemem: none, target_mem0: none, target_mem1: none) uwtable
+define dso_local void @foo(i32 noundef %n, i32 noundef %m, i32 noundef %p) local_unnamed_addr #0 {
+; CHECK-LABEL: @foo(
+; CHECK-NEXT:  entry:
+; CHECK-NEXT:    br label [[FOR_COND1_PREHEADER:%.*]]
+; CHECK:       for.cond1.preheader:
+; CHECK-NEXT:    [[I_058:%.*]] = phi i32 [ 0, [[ENTRY:%.*]] ], [ [[INC38:%.*]], [[FOR_INC37:%.*]] ]
+; CHECK-NEXT:    br label [[FOR_COND4_PREHEADER:%.*]]
+; CHECK:       for.cond4.preheader:
+; CHECK-NEXT:    [[INDVARS_IV60:%.*]] = phi i64 [ 0, [[FOR_COND1_PREHEADER]] ], [ [[INDVARS_IV_NEXT61:%.*]], [[FOR_INC13:%.*]] ]
+; CHECK-NEXT:    [[ARRAYIDX:%.*]] = getelementptr inbounds nuw [64 x i32], ptr @B, i64 [[INDVARS_IV60]]
+; CHECK-NEXT:    [[ARRAYIDX10:%.*]] = getelementptr inbounds nuw [64 x i32], ptr @C, i64 [[INDVARS_IV60]]
+; CHECK-NEXT:    br label [[FOR_BODY6:%.*]]
+; CHECK:       for.body6:
+; CHECK-NEXT:    [[INDVARS_IV:%.*]] = phi i64 [ 0, [[FOR_COND4_PREHEADER]] ], [ [[INDVARS_IV_NEXT:%.*]], [[FOR_BODY6]] ]
+; CHECK-NEXT:    [[ARRAYIDX8:%.*]] = getelementptr inbounds nuw i32, ptr [[ARRAYIDX]], i64 [[INDVARS_IV]]
+; CHECK-NEXT:    [[TMP0:%.*]] = load i32, ptr [[ARRAYIDX8]], align 4, !tbaa [[TBAA5:![0-9]+]]
+; CHECK-NEXT:    [[ADD:%.*]] = add nsw i32 [[TMP0]], [[I_058]]
+; CHECK-NEXT:    [[ARRAYIDX12:%.*]] = getelementptr inbounds nuw i32, ptr [[ARRAYIDX10]], i64 [[INDVARS_IV]]
+; CHECK-NEXT:    store i32 [[ADD]], ptr [[ARRAYIDX12]], align 4, !tbaa [[TBAA5]]
+; CHECK-NEXT:    [[INDVARS_IV_NEXT]] = add nuw nsw i64 [[INDVARS_IV]], 1
+; CHECK-NEXT:    [[EXITCOND_NOT:%.*]] = icmp eq i64 [[INDVARS_IV_NEXT]], 64
+; CHECK-NEXT:    br i1 [[EXITCOND_NOT]], label [[FOR_INC13]], label [[FOR_BODY6]], !llvm.loop [[LOOP9:![0-9]+]]
+; CHECK:       for.inc13:
+; CHECK-NEXT:    [[INDVARS_IV_NEXT61]] = add nuw nsw i64 [[INDVARS_IV60]], 1
+; CHECK-NEXT:    [[EXITCOND63_NOT:%.*]] = icmp eq i64 [[INDVARS_IV_NEXT61]], 64
+; CHECK-NEXT:    br i1 [[EXITCOND63_NOT]], label [[FOR_BODY21_PREHEADER:%.*]], label [[FOR_COND4_PREHEADER]], !llvm.loop [[LOOP12:![0-9]+]]
+; CHECK:       for.cond19.preheader.preheader:
+; CHECK-NEXT:    br label [[FOR_COND19_PREHEADER:%.*]]
+; CHECK:       for.cond19.preheader:
+; CHECK-NEXT:    [[INDVARS_IV68:%.*]] = phi i64 [ [[INDVARS_IV_NEXT69:%.*]], [[FOR_INC34:%.*]] ], [ 0, [[FOR_COND19_PREHEADER_PREHEADER:%.*]] ]
+; CHECK-NEXT:    [[INVARIANT_GEP:%.*]] = getelementptr inbounds nuw i32, ptr @A, i64 [[INDVARS_IV68]]
+; CHECK-NEXT:    br label [[FOR_BODY21_SPLIT1:%.*]]
+; CHECK:       for.body21.preheader:
+; CHECK-NEXT:    br label [[FOR_BODY21:%.*]]
+; CHECK:       for.body21:
+; CHECK-NEXT:    [[INDVARS_IV64:%.*]] = phi i64 [ [[TMP2:%.*]], [[FOR_BODY21_SPLIT:%.*]] ], [ 0, [[FOR_BODY21_PREHEADER]] ]
+; CHECK-NEXT:    br label [[FOR_COND19_PREHEADER_PREHEADER]]
+; CHECK:       for.body21.split1:
+; CHECK-NEXT:    [[GEP:%.*]] = getelementptr inbounds nuw [64 x i32], ptr [[INVARIANT_GEP]], i64 [[INDVARS_IV64]]
+; CHECK-NEXT:    [[TMP1:%.*]] = load i32, ptr [[GEP]], align 4, !tbaa [[TBAA5]]
+; CHECK-NEXT:    [[ADD26:%.*]] = add nsw i32 [[TMP1]], 1
+; CHECK-NEXT:    store i32 [[ADD26]], ptr [[GEP]], align 4, !tbaa [[TBAA5]]
+; CHECK-NEXT:    [[INDVARS_IV_NEXT65:%.*]] = add nuw nsw i64 [[INDVARS_IV64]], 1
+; CHECK-NEXT:    [[EXITCOND67_NOT:%.*]] = icmp eq i64 [[INDVARS_IV_NEXT65]], 64
+; CHECK-NEXT:    br label [[FOR_INC34]]
+; CHECK:       for.body21.split:
+; CHECK-NEXT:    [[TMP2]] = add nuw nsw i64 [[INDVARS_IV64]], 1
+; CHECK-NEXT:    [[TMP3:%.*]] = icmp eq i64 [[TMP2]], 64
+; CHECK-NEXT:    br i1 [[TMP3]], label [[FOR_INC37]], label [[FOR_BODY21]], !llvm.loop [[LOOP13:![0-9]+]]
+; CHECK:       for.inc34:
+; CHECK-NEXT:    [[INDVARS_IV_NEXT69]] = add nuw nsw i64 [[INDVARS_IV68]], 1
+; CHECK-NEXT:    [[EXITCOND71_NOT:%.*]] = icmp eq i64 [[INDVARS_IV_NEXT69]], 64
+; CHECK-NEXT:    br i1 [[EXITCOND71_NOT]], label [[FOR_BODY21_SPLIT]], label [[FOR_COND19_PREHEADER]], !llvm.loop [[LOOP14:![0-9]+]]
+; CHECK:       for.inc37:
+; CHECK-NEXT:    [[INC38]] = add nuw nsw i32 [[I_058]], 1
+; CHECK-NEXT:    [[EXITCOND72_NOT:%.*]] = icmp eq i32 [[INC38]], 64
+; CHECK-NEXT:    br i1 [[EXITCOND72_NOT]], label [[FOR_END39:%.*]], label [[FOR_COND1_PREHEADER]], !llvm.loop [[LOOP15:![0-9]+]]
+; CHECK:       for.end39:
+; CHECK-NEXT:    ret void
+;
+entry:
+  br label %for.cond1.preheader
+
+for.cond1.preheader:                              ; preds = %entry, %for.inc37
+  %i.058 = phi i32 [ 0, %entry ], [ %inc38, %for.inc37 ]
+  br label %for.cond4.preheader
+
+for.cond4.preheader:                              ; preds = %for.cond1.preheader, %for.inc13
+  %indvars.iv60 = phi i64 [ 0, %for.cond1.preheader ], [ %indvars.iv.next61, %for.inc13 ]
+  %arrayidx = getelementptr inbounds nuw [64 x i32], ptr @B, i64 %indvars.iv60
+  %arrayidx10 = getelementptr inbounds nuw [64 x i32], ptr @C, i64 %indvars.iv60
+  br label %for.body6
+
+for.body6:                                        ; preds = %for.cond4.preheader, %for.body6
+  %indvars.iv = phi i64 [ 0, %for.cond4.preheader ], [ %indvars.iv.next, %for.body6 ]
+  %arrayidx8 = getelementptr inbounds nuw i32, ptr %arrayidx, i64 %indvars.iv
+  %0 = load i32, ptr %arrayidx8, align 4, !tbaa !5
+  %add = add nsw i32 %0, %i.058
+  %arrayidx12 = getelementptr inbounds nuw i32, ptr %arrayidx10, i64 %indvars.iv
+  store i32 %add, ptr %arrayidx12, align 4, !tbaa !5
+  %indvars.iv.next = add nuw nsw i64 %indvars.iv, 1
+  %exitcond.not = icmp eq i64 %indvars.iv.next, 64
+  br i1 %exitcond.not, label %for.inc13, label %for.body6, !llvm.loop !9
+
+for.inc13:                                        ; preds = %for.body6
+  %indvars.iv.next61 = add nuw nsw i64 %indvars.iv60, 1
+  %exitcond63.not = icmp eq i64 %indvars.iv.next61, 64
+  br i1 %exitcond63.not, label %for.cond19.preheader.preheader, label %for.cond4.preheader, !llvm.loop !12
+
+for.cond19.preheader.preheader:                   ; preds = %for.inc13
+  br label %for.cond19.preheader
+
+for.cond19.preheader:                             ; preds = %for.cond19.preheader.preheader, %for.inc34
+  %indvars.iv68 = phi i64 [ %indvars.iv.next69, %for.inc34 ], [ 0, %for.cond19.preheader.preheader ]
+  %invariant.gep = getelementptr inbounds nuw i32, ptr @A, i64 %indvars.iv68
+  br label %for.body21
+
+for.body21:                                       ; preds = %for.cond19.preheader, %for.body21
+  %indvars.iv64 = phi i64 [ 0, %for.cond19.preheader ], [ %indvars.iv.next65, %for.body21 ]
+  %gep = getelementptr inbounds nuw [64 x i32], ptr %invariant.gep, i64 %indvars.iv64
+  %1 = load i32, ptr %gep, align 4, !tbaa !5
+  %add26 = add nsw i32 %1, 1
+  store i32 %add26, ptr %gep, align 4, !tbaa !5
+  %indvars.iv.next65 = add nuw nsw i64 %indvars.iv64, 1
+  %exitcond67.not = icmp eq i64 %indvars.iv.next65, 64
+  br i1 %exitcond67.not, label %for.inc34, label %for.body21, !llvm.loop !13
+
+for.inc34:                                        ; preds = %for.body21
+  %indvars.iv.next69 = add nuw nsw i64 %indvars.iv68, 1
+  %exitcond71.not = icmp eq i64 %indvars.iv.next69, 64
+  br i1 %exitcond71.not, label %for.inc37, label %for.cond19.preheader, !llvm.loop !14
+
+for.inc37:                                        ; preds = %for.inc34
+  %inc38 = add nuw nsw i32 %i.058, 1
+  %exitcond72.not = icmp eq i32 %inc38, 64
+  br i1 %exitcond72.not, label %for.end39, label %for.cond1.preheader, !llvm.loop !15
+
+for.end39:                                        ; preds = %for.inc37
+  ret void
+}
+
+attributes #0 = { nofree norecurse nosync nounwind memory(readwrite, argmem: none, inaccessiblemem: none, target_mem0: none, target_mem1: none) uwtable "frame-pointer"="non-leaf-no-reserve" "no-trapping-math"="true" "stack-protector-buffer-size"="8" "target-cpu"="generic" "target-features"="+fp-armv8,+neon,+v8a,-fmv" }
+
+!llvm.module.flags = !{!0, !1, !2, !3}
+!llvm.ident = !{!4}
+!llvm.errno.tbaa = !{!5}
+
+!0 = !{i32 8, !"PIC Level", i32 2}
+!1 = !{i32 7, !"PIE Level", i32 2}
+!2 = !{i32 7, !"uwtable", i32 2}
+!3 = !{i32 7, !"frame-pointer", i32 4}
+!4 = !{!"clang version 23.0.0git (ssh://git-hexagon-lv.quicinc.com:29418/llvm/llvm-project 0b3252c0b4f1cb5394a6db1eba0f70190fc88827)"}
+!5 = !{!6, !6, i64 0}
+!6 = !{!"int", !7, i64 0}
+!7 = !{!"omnipotent char", !8, i64 0}
+!8 = !{!"Simple C/C++ TBAA"}
+!9 = distinct !{!9, !10, !11}
+!10 = !{!"llvm.loop.mustprogress"}
+!11 = !{!"llvm.loop.unroll.disable"}
+!12 = distinct !{!12, !10}
+!13 = distinct !{!13, !10, !11}
+!14 = distinct !{!14, !10}
+!15 = distinct !{!15, !10}
diff --git a/llvm/test/Transforms/LoopInterchange/large-nested-6d.ll b/llvm/test/Transforms/LoopInterchange/large-nested-6d.ll
index 590c21fd5a1be..dfe4857b52485 100644
--- a/llvm/test/Transforms/LoopInterchange/large-nested-6d.ll
+++ b/llvm/test/Transforms/LoopInterchange/large-nested-6d.ll
@@ -1,5 +1,6 @@
 ; RUN: opt < %s -passes=loop-interchange -cache-line-size=64 -pass-remarks='loop-interchange' -pass-remarks-missed='loop-interchange' -pass-remarks-output=%t -disable-output -S
 ; RUN: FileCheck --input-file=%t %s
+; XFAIL: *
 
 target datalayout = "e-m:e-p270:32:32-p271:32:32-p272:64:64-i8:8:32-i16:16:32-i64:64-i128:128-n32:64-S128-Fn32"
 

>From 3a6ac775ad0de26d145fefee0cc0af8ef1a68380 Mon Sep 17 00:00:00 2001
From: rohgarg <rohgarg at qti.qualcomm.com>
Date: Tue, 9 Jun 2026 12:41:51 -0700
Subject: [PATCH 2/9] [LoopInterchange] Collect perfect subnests bottom-up from
 leaf loops

---
 .../lib/Transforms/Scalar/LoopInterchange.cpp | 114 +++++++++---------
 1 file changed, 60 insertions(+), 54 deletions(-)

diff --git a/llvm/test/Transforms/LoopInterchange/bail-out-one-loop.ll b/llvm/test/Transforms/LoopInterchange/bail-out-one-loop.ll
index d1cf33acd2831..efe25f8455655 100644
--- a/llvm/test/Transforms/LoopInterchange/bail-out-one-loop.ll
+++ b/llvm/test/Transforms/LoopInterchange/bail-out-one-loop.ll
@@ -1,6 +1,7 @@
 ; REQUIRES: asserts
 
 ; RUN: opt < %s -passes=loop-interchange -debug -disable-output 2>&1 | FileCheck %s
+; XFAIL: *
 
 target datalayout = "e-m:e-p270:32:32-p271:32:32-p272:64:64-i8:8:32-i16:16:32-i64:64-i128:128-n32:64-S128-Fn32"
 
diff --git a/llvm/test/Transforms/LoopInterchange/imperfect-loop-nest.ll b/llvm/test/Transforms/LoopInterchange/imperfect-loop-nest.ll
deleted file mode 100644
index d552d32e46377..0000000000000
--- a/llvm/test/Transforms/LoopInterchange/imperfect-loop-nest.ll
+++ /dev/null
@@ -1,176 +0,0 @@
-; NOTE: Assertions have been autogenerated by utils/update_test_checks.py
-; RUN: opt -S -passes='loop-interchange' -cache-line-size=64 < %s | FileCheck %s
-
-;  // Loop nest with multiple loops at same depth
-;    for (i = 0; i < 64; i++)
-;    {
-;      for (j = 0; j < 64; j++)
-;        {
-;          #pragma clang loop unroll(disable)
-;          for (r = 0; r < 64; r++)
-;            {
-;              C[j][r] = B[j][r] + i;
-;            }
-;        }
-;
-;      for (k = 0; k < 64; k++)
-;        {
-;          #pragma clang loop unroll(disable)
-;          for (l = 0; l < 64; l++)
-;            {
-;              A[l][k] = A[l][k] + 1;
-;            }
-;        }
-;    }
-
-target datalayout = "e-m:e-p270:32:32-p271:32:32-p272:64:64-i8:8:32-i16:16:32-i64:64-i128:128-n32:64-S128-Fn32"
-target triple = "aarch64-unknown-linux-gnueabi"
-
- at B = dso_local local_unnamed_addr global [64 x [64 x i32]] zeroinitializer, align 4
- at C = dso_local local_unnamed_addr global [64 x [64 x i32]] zeroinitializer, align 4
- at A = dso_local local_unnamed_addr global [64 x [64 x i32]] zeroinitializer, align 4
-
-; Function Attrs: nofree norecurse nosync nounwind memory(readwrite, argmem: none, inaccessiblemem: none, target_mem0: none, target_mem1: none) uwtable
-define dso_local void @foo(i32 noundef %n, i32 noundef %m, i32 noundef %p) local_unnamed_addr #0 {
-; CHECK-LABEL: @foo(
-; CHECK-NEXT:  entry:
-; CHECK-NEXT:    br label [[FOR_COND1_PREHEADER:%.*]]
-; CHECK:       for.cond1.preheader:
-; CHECK-NEXT:    [[I_058:%.*]] = phi i32 [ 0, [[ENTRY:%.*]] ], [ [[INC38:%.*]], [[FOR_INC37:%.*]] ]
-; CHECK-NEXT:    br label [[FOR_COND4_PREHEADER:%.*]]
-; CHECK:       for.cond4.preheader:
-; CHECK-NEXT:    [[INDVARS_IV60:%.*]] = phi i64 [ 0, [[FOR_COND1_PREHEADER]] ], [ [[INDVARS_IV_NEXT61:%.*]], [[FOR_INC13:%.*]] ]
-; CHECK-NEXT:    [[ARRAYIDX:%.*]] = getelementptr inbounds nuw [64 x i32], ptr @B, i64 [[INDVARS_IV60]]
-; CHECK-NEXT:    [[ARRAYIDX10:%.*]] = getelementptr inbounds nuw [64 x i32], ptr @C, i64 [[INDVARS_IV60]]
-; CHECK-NEXT:    br label [[FOR_BODY6:%.*]]
-; CHECK:       for.body6:
-; CHECK-NEXT:    [[INDVARS_IV:%.*]] = phi i64 [ 0, [[FOR_COND4_PREHEADER]] ], [ [[INDVARS_IV_NEXT:%.*]], [[FOR_BODY6]] ]
-; CHECK-NEXT:    [[ARRAYIDX8:%.*]] = getelementptr inbounds nuw i32, ptr [[ARRAYIDX]], i64 [[INDVARS_IV]]
-; CHECK-NEXT:    [[TMP0:%.*]] = load i32, ptr [[ARRAYIDX8]], align 4, !tbaa [[TBAA5:![0-9]+]]
-; CHECK-NEXT:    [[ADD:%.*]] = add nsw i32 [[TMP0]], [[I_058]]
-; CHECK-NEXT:    [[ARRAYIDX12:%.*]] = getelementptr inbounds nuw i32, ptr [[ARRAYIDX10]], i64 [[INDVARS_IV]]
-; CHECK-NEXT:    store i32 [[ADD]], ptr [[ARRAYIDX12]], align 4, !tbaa [[TBAA5]]
-; CHECK-NEXT:    [[INDVARS_IV_NEXT]] = add nuw nsw i64 [[INDVARS_IV]], 1
-; CHECK-NEXT:    [[EXITCOND_NOT:%.*]] = icmp eq i64 [[INDVARS_IV_NEXT]], 64
-; CHECK-NEXT:    br i1 [[EXITCOND_NOT]], label [[FOR_INC13]], label [[FOR_BODY6]], !llvm.loop [[LOOP9:![0-9]+]]
-; CHECK:       for.inc13:
-; CHECK-NEXT:    [[INDVARS_IV_NEXT61]] = add nuw nsw i64 [[INDVARS_IV60]], 1
-; CHECK-NEXT:    [[EXITCOND63_NOT:%.*]] = icmp eq i64 [[INDVARS_IV_NEXT61]], 64
-; CHECK-NEXT:    br i1 [[EXITCOND63_NOT]], label [[FOR_BODY21_PREHEADER:%.*]], label [[FOR_COND4_PREHEADER]], !llvm.loop [[LOOP12:![0-9]+]]
-; CHECK:       for.cond19.preheader.preheader:
-; CHECK-NEXT:    br label [[FOR_COND19_PREHEADER:%.*]]
-; CHECK:       for.cond19.preheader:
-; CHECK-NEXT:    [[INDVARS_IV68:%.*]] = phi i64 [ [[INDVARS_IV_NEXT69:%.*]], [[FOR_INC34:%.*]] ], [ 0, [[FOR_COND19_PREHEADER_PREHEADER:%.*]] ]
-; CHECK-NEXT:    [[INVARIANT_GEP:%.*]] = getelementptr inbounds nuw i32, ptr @A, i64 [[INDVARS_IV68]]
-; CHECK-NEXT:    br label [[FOR_BODY21_SPLIT1:%.*]]
-; CHECK:       for.body21.preheader:
-; CHECK-NEXT:    br label [[FOR_BODY21:%.*]]
-; CHECK:       for.body21:
-; CHECK-NEXT:    [[INDVARS_IV64:%.*]] = phi i64 [ [[TMP2:%.*]], [[FOR_BODY21_SPLIT:%.*]] ], [ 0, [[FOR_BODY21_PREHEADER]] ]
-; CHECK-NEXT:    br label [[FOR_COND19_PREHEADER_PREHEADER]]
-; CHECK:       for.body21.split1:
-; CHECK-NEXT:    [[GEP:%.*]] = getelementptr inbounds nuw [64 x i32], ptr [[INVARIANT_GEP]], i64 [[INDVARS_IV64]]
-; CHECK-NEXT:    [[TMP1:%.*]] = load i32, ptr [[GEP]], align 4, !tbaa [[TBAA5]]
-; CHECK-NEXT:    [[ADD26:%.*]] = add nsw i32 [[TMP1]], 1
-; CHECK-NEXT:    store i32 [[ADD26]], ptr [[GEP]], align 4, !tbaa [[TBAA5]]
-; CHECK-NEXT:    [[INDVARS_IV_NEXT65:%.*]] = add nuw nsw i64 [[INDVARS_IV64]], 1
-; CHECK-NEXT:    [[EXITCOND67_NOT:%.*]] = icmp eq i64 [[INDVARS_IV_NEXT65]], 64
-; CHECK-NEXT:    br label [[FOR_INC34]]
-; CHECK:       for.body21.split:
-; CHECK-NEXT:    [[TMP2]] = add nuw nsw i64 [[INDVARS_IV64]], 1
-; CHECK-NEXT:    [[TMP3:%.*]] = icmp eq i64 [[TMP2]], 64
-; CHECK-NEXT:    br i1 [[TMP3]], label [[FOR_INC37]], label [[FOR_BODY21]], !llvm.loop [[LOOP13:![0-9]+]]
-; CHECK:       for.inc34:
-; CHECK-NEXT:    [[INDVARS_IV_NEXT69]] = add nuw nsw i64 [[INDVARS_IV68]], 1
-; CHECK-NEXT:    [[EXITCOND71_NOT:%.*]] = icmp eq i64 [[INDVARS_IV_NEXT69]], 64
-; CHECK-NEXT:    br i1 [[EXITCOND71_NOT]], label [[FOR_BODY21_SPLIT]], label [[FOR_COND19_PREHEADER]], !llvm.loop [[LOOP14:![0-9]+]]
-; CHECK:       for.inc37:
-; CHECK-NEXT:    [[INC38]] = add nuw nsw i32 [[I_058]], 1
-; CHECK-NEXT:    [[EXITCOND72_NOT:%.*]] = icmp eq i32 [[INC38]], 64
-; CHECK-NEXT:    br i1 [[EXITCOND72_NOT]], label [[FOR_END39:%.*]], label [[FOR_COND1_PREHEADER]], !llvm.loop [[LOOP15:![0-9]+]]
-; CHECK:       for.end39:
-; CHECK-NEXT:    ret void
-;
-entry:
-  br label %for.cond1.preheader
-
-for.cond1.preheader:                              ; preds = %entry, %for.inc37
-  %i.058 = phi i32 [ 0, %entry ], [ %inc38, %for.inc37 ]
-  br label %for.cond4.preheader
-
-for.cond4.preheader:                              ; preds = %for.cond1.preheader, %for.inc13
-  %indvars.iv60 = phi i64 [ 0, %for.cond1.preheader ], [ %indvars.iv.next61, %for.inc13 ]
-  %arrayidx = getelementptr inbounds nuw [64 x i32], ptr @B, i64 %indvars.iv60
-  %arrayidx10 = getelementptr inbounds nuw [64 x i32], ptr @C, i64 %indvars.iv60
-  br label %for.body6
-
-for.body6:                                        ; preds = %for.cond4.preheader, %for.body6
-  %indvars.iv = phi i64 [ 0, %for.cond4.preheader ], [ %indvars.iv.next, %for.body6 ]
-  %arrayidx8 = getelementptr inbounds nuw i32, ptr %arrayidx, i64 %indvars.iv
-  %0 = load i32, ptr %arrayidx8, align 4, !tbaa !5
-  %add = add nsw i32 %0, %i.058
-  %arrayidx12 = getelementptr inbounds nuw i32, ptr %arrayidx10, i64 %indvars.iv
-  store i32 %add, ptr %arrayidx12, align 4, !tbaa !5
-  %indvars.iv.next = add nuw nsw i64 %indvars.iv, 1
-  %exitcond.not = icmp eq i64 %indvars.iv.next, 64
-  br i1 %exitcond.not, label %for.inc13, label %for.body6, !llvm.loop !9
-
-for.inc13:                                        ; preds = %for.body6
-  %indvars.iv.next61 = add nuw nsw i64 %indvars.iv60, 1
-  %exitcond63.not = icmp eq i64 %indvars.iv.next61, 64
-  br i1 %exitcond63.not, label %for.cond19.preheader.preheader, label %for.cond4.preheader, !llvm.loop !12
-
-for.cond19.preheader.preheader:                   ; preds = %for.inc13
-  br label %for.cond19.preheader
-
-for.cond19.preheader:                             ; preds = %for.cond19.preheader.preheader, %for.inc34
-  %indvars.iv68 = phi i64 [ %indvars.iv.next69, %for.inc34 ], [ 0, %for.cond19.preheader.preheader ]
-  %invariant.gep = getelementptr inbounds nuw i32, ptr @A, i64 %indvars.iv68
-  br label %for.body21
-
-for.body21:                                       ; preds = %for.cond19.preheader, %for.body21
-  %indvars.iv64 = phi i64 [ 0, %for.cond19.preheader ], [ %indvars.iv.next65, %for.body21 ]
-  %gep = getelementptr inbounds nuw [64 x i32], ptr %invariant.gep, i64 %indvars.iv64
-  %1 = load i32, ptr %gep, align 4, !tbaa !5
-  %add26 = add nsw i32 %1, 1
-  store i32 %add26, ptr %gep, align 4, !tbaa !5
-  %indvars.iv.next65 = add nuw nsw i64 %indvars.iv64, 1
-  %exitcond67.not = icmp eq i64 %indvars.iv.next65, 64
-  br i1 %exitcond67.not, label %for.inc34, label %for.body21, !llvm.loop !13
-
-for.inc34:                                        ; preds = %for.body21
-  %indvars.iv.next69 = add nuw nsw i64 %indvars.iv68, 1
-  %exitcond71.not = icmp eq i64 %indvars.iv.next69, 64
-  br i1 %exitcond71.not, label %for.inc37, label %for.cond19.preheader, !llvm.loop !14
-
-for.inc37:                                        ; preds = %for.inc34
-  %inc38 = add nuw nsw i32 %i.058, 1
-  %exitcond72.not = icmp eq i32 %inc38, 64
-  br i1 %exitcond72.not, label %for.end39, label %for.cond1.preheader, !llvm.loop !15
-
-for.end39:                                        ; preds = %for.inc37
-  ret void
-}
-
-attributes #0 = { nofree norecurse nosync nounwind memory(readwrite, argmem: none, inaccessiblemem: none, target_mem0: none, target_mem1: none) uwtable "frame-pointer"="non-leaf-no-reserve" "no-trapping-math"="true" "stack-protector-buffer-size"="8" "target-cpu"="generic" "target-features"="+fp-armv8,+neon,+v8a,-fmv" }
-
-!llvm.module.flags = !{!0, !1, !2, !3}
-!llvm.ident = !{!4}
-!llvm.errno.tbaa = !{!5}
-
-!0 = !{i32 8, !"PIC Level", i32 2}
-!1 = !{i32 7, !"PIE Level", i32 2}
-!2 = !{i32 7, !"uwtable", i32 2}
-!3 = !{i32 7, !"frame-pointer", i32 4}
-!4 = !{!"clang version 23.0.0git (ssh://git-hexagon-lv.quicinc.com:29418/llvm/llvm-project 0b3252c0b4f1cb5394a6db1eba0f70190fc88827)"}
-!5 = !{!6, !6, i64 0}
-!6 = !{!"int", !7, i64 0}
-!7 = !{!"omnipotent char", !8, i64 0}
-!8 = !{!"Simple C/C++ TBAA"}
-!9 = distinct !{!9, !10, !11}
-!10 = !{!"llvm.loop.mustprogress"}
-!11 = !{!"llvm.loop.unroll.disable"}
-!12 = distinct !{!12, !10}
-!13 = distinct !{!13, !10, !11}
-!14 = distinct !{!14, !10}
-!15 = distinct !{!15, !10}
diff --git a/llvm/test/Transforms/LoopInterchange/partially-perfect-loop.ll b/llvm/test/Transforms/LoopInterchange/partially-perfect-loop.ll
index 7d72b4a7f1804..89e63a3f1edaf 100644
--- a/llvm/test/Transforms/LoopInterchange/partially-perfect-loop.ll
+++ b/llvm/test/Transforms/LoopInterchange/partially-perfect-loop.ll
@@ -17,6 +17,7 @@ define void @f(ptr noalias %A, ptr noalias %B) {
 ; CHECK-NEXT:  [[FOR_J_HEADER:.*]]:
 ; CHECK-NEXT:    br label %[[FOR_R_HEADER_SPLIT1:.*]]
 ; CHECK:       [[FOR_R_HEADER_SPLIT1]]:
+<<<<<<< HEAD
 ; CHECK-NEXT:    [[I:%.*]] = phi i64 [ 0, %[[FOR_J_HEADER]] ], [ [[I_NEXT:%.*]], %[[FOR_I_LATCH:.*]] ]
 ; CHECK-NEXT:    br label %[[FOR_R_HEADER:.*]]
 ; CHECK:       [[FOR_R_HEADER]]:
@@ -34,10 +35,40 @@ define void @f(ptr noalias %A, ptr noalias %B) {
 ; CHECK-NEXT:    [[J_DONE:%.*]] = icmp eq i64 [[J_NEXT]], 64
 ; CHECK-NEXT:    br i1 [[J_DONE]], label %[[FOR_L_HEADER_PREHEADER:.*]], label %[[FOR_R_HEADER]]
 ; CHECK:       [[FOR_L_HEADER_PREHEADER]]:
+=======
+; CHECK-NEXT:    [[I:%.*]] = phi i64 [ 0, %[[FOR_J_HEADER]] ], [ [[I_NEXT:%.*]], %[[FOR_I_LATCH1:.*]] ]
+; CHECK-NEXT:    br label %[[FOR_R_HEADER:.*]]
+; CHECK:       [[FOR_J_HEADER_PREHEADER1:.*]]:
+; CHECK-NEXT:    br label %[[FOR_J_HEADER_PREHEADER:.*]]
+; CHECK:       [[FOR_J_HEADER_PREHEADER]]:
+; CHECK-NEXT:    [[J:%.*]] = phi i64 [ [[J_NEXT:%.*]], %[[FOR_J_LATCH1:.*]] ], [ 0, %[[FOR_J_HEADER_PREHEADER1]] ]
+; CHECK-NEXT:    br label %[[FOR_R_HEADER_SPLIT2:.*]]
+; CHECK:       [[FOR_R_HEADER]]:
+; CHECK-NEXT:    br label %[[FOR_R_HEADER1:.*]]
+; CHECK:       [[FOR_R_HEADER1]]:
+; CHECK-NEXT:    [[R:%.*]] = phi i64 [ [[TMP4:%.*]], %[[FOR_J_LATCH:.*]] ], [ 0, %[[FOR_R_HEADER]] ]
+; CHECK-NEXT:    br label %[[FOR_J_HEADER_PREHEADER1]]
+; CHECK:       [[FOR_R_HEADER_SPLIT2]]:
+; CHECK-NEXT:    [[A_ELEMENT:%.*]] = getelementptr [64 x i8], ptr [[A]], i64 [[J]], i64 [[R]]
+; CHECK-NEXT:    store i8 0, ptr [[A_ELEMENT]], align 1
+; CHECK-NEXT:    [[TMP0:%.*]] = add i64 [[R]], 1
+; CHECK-NEXT:    [[TMP1:%.*]] = icmp eq i64 [[TMP0]], 64
+; CHECK-NEXT:    br label %[[FOR_J_LATCH1]]
+; CHECK:       [[FOR_J_LATCH]]:
+; CHECK-NEXT:    [[TMP4]] = add i64 [[R]], 1
+; CHECK-NEXT:    [[TMP5:%.*]] = icmp eq i64 [[TMP4]], 64
+; CHECK-NEXT:    br i1 [[TMP5]], label %[[FOR_L_HEADER_PREHEADER1:.*]], label %[[FOR_R_HEADER1]]
+; CHECK:       [[FOR_J_LATCH1]]:
+; CHECK-NEXT:    [[J_NEXT]] = add i64 [[J]], 1
+; CHECK-NEXT:    [[J_DONE:%.*]] = icmp eq i64 [[J_NEXT]], 64
+; CHECK-NEXT:    br i1 [[J_DONE]], label %[[FOR_J_LATCH]], label %[[FOR_J_HEADER_PREHEADER]]
+; CHECK:       [[FOR_L_HEADER_PREHEADER:.*]]:
+>>>>>>> effa84a158f2 ([LoopInterchange] Updated the tests for imperfect-loop-nests)
 ; CHECK-NEXT:    br label %[[FOR_L_HEADER:.*]]
 ; CHECK:       [[FOR_L_HEADER]]:
 ; CHECK-NEXT:    [[K:%.*]] = phi i64 [ [[K_NEXT:%.*]], %[[FOR_K_LATCH:.*]] ], [ 0, %[[FOR_L_HEADER_PREHEADER]] ]
 ; CHECK-NEXT:    br label %[[FOR_K_HEADER_PREHEADER:.*]]
+<<<<<<< HEAD
 ; CHECK:       [[FOR_K_HEADER_PREHEADER]]:
 ; CHECK-NEXT:    [[L:%.*]] = phi i64 [ 0, %[[FOR_L_HEADER]] ], [ [[TMP2:%.*]], %[[FOR_K_HEADER_PREHEADER]] ]
 ; CHECK-NEXT:    [[B_ELEMENT:%.*]] = getelementptr [64 x i8], ptr [[B]], i64 [[L]], i64 [[K]]
@@ -45,11 +76,32 @@ define void @f(ptr noalias %A, ptr noalias %B) {
 ; CHECK-NEXT:    [[TMP2]] = add i64 [[L]], 1
 ; CHECK-NEXT:    [[TMP3:%.*]] = icmp eq i64 [[TMP2]], 64
 ; CHECK-NEXT:    br i1 [[TMP3]], label %[[FOR_K_LATCH]], label %[[FOR_K_HEADER_PREHEADER]]
+=======
+; CHECK:       [[FOR_L_HEADER_PREHEADER1]]:
+; CHECK-NEXT:    br label %[[FOR_L_HEADER1:.*]]
+; CHECK:       [[FOR_L_HEADER1]]:
+; CHECK-NEXT:    [[L:%.*]] = phi i64 [ [[TMP6:%.*]], %[[FOR_I_LATCH:.*]] ], [ 0, %[[FOR_L_HEADER_PREHEADER1]] ]
+; CHECK-NEXT:    br label %[[FOR_L_HEADER_PREHEADER]]
+; CHECK:       [[FOR_K_HEADER_PREHEADER]]:
+; CHECK-NEXT:    [[B_ELEMENT:%.*]] = getelementptr [64 x i8], ptr [[B]], i64 [[L]], i64 [[K]]
+; CHECK-NEXT:    store i8 0, ptr [[B_ELEMENT]], align 1
+; CHECK-NEXT:    [[TMP2:%.*]] = add i64 [[L]], 1
+; CHECK-NEXT:    [[TMP3:%.*]] = icmp eq i64 [[TMP2]], 64
+; CHECK-NEXT:    br label %[[FOR_K_LATCH]]
+; CHECK:       [[FOR_I_LATCH]]:
+; CHECK-NEXT:    [[TMP6]] = add i64 [[L]], 1
+; CHECK-NEXT:    [[TMP7:%.*]] = icmp eq i64 [[TMP6]], 64
+; CHECK-NEXT:    br i1 [[TMP7]], label %[[FOR_I_LATCH1]], label %[[FOR_L_HEADER1]]
+>>>>>>> effa84a158f2 ([LoopInterchange] Updated the tests for imperfect-loop-nests)
 ; CHECK:       [[FOR_K_LATCH]]:
 ; CHECK-NEXT:    [[K_NEXT]] = add i64 [[K]], 1
 ; CHECK-NEXT:    [[K_DONE:%.*]] = icmp eq i64 [[K_NEXT]], 64
 ; CHECK-NEXT:    br i1 [[K_DONE]], label %[[FOR_I_LATCH]], label %[[FOR_L_HEADER]]
+<<<<<<< HEAD
 ; CHECK:       [[FOR_I_LATCH]]:
+=======
+; CHECK:       [[FOR_I_LATCH1]]:
+>>>>>>> effa84a158f2 ([LoopInterchange] Updated the tests for imperfect-loop-nests)
 ; CHECK-NEXT:    [[I_NEXT]] = add i64 [[I]], 1
 ; CHECK-NEXT:    [[I_DONE:%.*]] = icmp eq i64 [[I_NEXT]], 64
 ; CHECK-NEXT:    br i1 [[I_DONE]], label %[[EXIT:.*]], label %[[FOR_R_HEADER_SPLIT1]]

>From cbf95543de7d1706fba3f299eabf04f058c331e5 Mon Sep 17 00:00:00 2001
From: rohgarg <rohgarg at qti.qualcomm.com>
Date: Mon, 15 Jun 2026 13:38:53 -0700
Subject: [PATCH 4/9] [LoopInterchange] Remove redundant check and update
 padding logic in dependency matrix construction

---
 .../lib/Transforms/Scalar/LoopInterchange.cpp | 64 ++++++++-----------
 .../LoopInterchange/partially-perfect-loop.ll | 34 ----------
 2 files changed, 28 insertions(+), 70 deletions(-)

diff --git a/llvm/lib/Transforms/Scalar/LoopInterchange.cpp b/llvm/lib/Transforms/Scalar/LoopInterchange.cpp
index 7617c674681ea..e76a20ee6beba 100644
--- a/llvm/lib/Transforms/Scalar/LoopInterchange.cpp
+++ b/llvm/lib/Transforms/Scalar/LoopInterchange.cpp
@@ -167,8 +167,7 @@ static bool inThisOrder(const Instruction *Src, const Instruction *Dst) {
 #endif
 
 static bool populateDependencyMatrix(CharMatrix &DepMatrix, unsigned Level,
-                                     Loop *L, ArrayRef<Loop *> LoopList,
-                                     LoopInfo *LI, DependenceInfo *DI,
+                                     Loop *L, DependenceInfo *DI,
                                      ScalarEvolution *SE,
                                      OptimizationRemarkEmitter *ORE) {
   using ValueVector = SmallVector<Value *, 16>;
@@ -176,14 +175,8 @@ static bool populateDependencyMatrix(CharMatrix &DepMatrix, unsigned Level,
   ValueVector MemInstr;
   unsigned NumInsts = 0;
 
-  // Collect memory instructions from the BB which belongs to all loops in the
-  // LoopList
-  for (BasicBlock *BB : L->getBlocksVector()) {
-    // In this iteration we need to handle the BB contained directly by Current
-    // Loop and all other BB of subloops will be handled in its own in next
-    // iterations.
-    if (!llvm::is_contained(LoopList, LI->getLoopFor(BB)))
-      continue;
+  // For each block.
+  for (BasicBlock *BB : L->blocks()) {
     // Scan the BB and collect legal loads and stores.
     for (Instruction &I : *BB) {
       NumInsts++;
@@ -239,9 +232,8 @@ static bool populateDependencyMatrix(CharMatrix &DepMatrix, unsigned Level,
         // make it non-negative.
         if (D->normalize(SE))
           LLVM_DEBUG(dbgs() << "Negative dependence vector normalized.\n");
-        LLVM_DEBUG(StringRef DepType = D->isFlow()   ? "flow"
-                                       : D->isAnti() ? "anti"
-                                                     : "output";
+        LLVM_DEBUG(StringRef DepType =
+                       D->isFlow() ? "flow" : D->isAnti() ? "anti" : "output";
                    dbgs() << "Found " << DepType
                           << " dependency between Src and Dst\n"
                           << " Src:" << *Src << "\n Dst:" << *Dst << '\n');
@@ -273,7 +265,7 @@ static bool populateDependencyMatrix(CharMatrix &DepMatrix, unsigned Level,
           Dep.assign(Level, '*');
         }
 
-        while (Dep.size() < Level) {
+        while (Dep.size() < L->getLoopDepth() + Level - 1) {
           Dep.push_back('I');
         }
 
@@ -688,6 +680,22 @@ struct LoopInterchange {
     return processLoopList(LoopList);
   }
 
+  /// Consider below kernel:
+  /// for(int i=0; i<n; i++){  // Loop 1
+  ///     for(int j=0; j<m; j++){  // Loop 2
+  ///         for(int r=0; r<m; r++){  // Loop 3
+  ///         // Do something
+  ///         }
+  ///     }
+  ///     for(int k=0; k<p; k++){  // Loop 4
+  ///         for(int l=0; l<p; l++){  // Loop 5
+  ///             // Do something
+  ///         }
+  ///     }
+  /// }
+  /// Then collectPerfectNests() will return:
+  /// - [Loop2, Loop3]
+  /// - [Loop4, Loop5]
   static SmallVector<SmallVector<Loop *, 8>, 4>
   collectPerfectNests(LoopNest &LN) {
     SmallVector<SmallVector<Loop *, 8>, 4> LoopLists;
@@ -713,22 +721,6 @@ struct LoopInterchange {
 
   bool run(LoopNest &LN) {
     SmallVector<SmallVector<Loop *, 8>, 4> LoopLists = collectPerfectNests(LN);
-    // Consider below kernel
-    // for(int i=0; i<n; i++){  // Loop 1
-    //     for(int j=0; j<m; j++){  // Loop 2
-    //         for(int r=0; r<m; r++){  // Loop 3
-    //         // Do something
-    //         }
-    //     }
-    //     for(int k=0; k<p; k++){  // Loop 4
-    //         for(int l=0; l<p; l++){  // Loop 5
-    //             // Do something
-    //         }
-    //     }
-    // }
-    // Then LoopLists will contain:
-    // - [Loop2, Loop3]
-    // - [Loop4, Loop5]
     bool Changed = false;
     for (SmallVector<Loop *, 8> &LoopList : LoopLists) {
       // Ensure minimum depth of the loop nest to do the interchange.
@@ -771,7 +763,7 @@ struct LoopInterchange {
     CharMatrix DependencyMatrix;
     Loop *OuterMostLoop = *(LoopList.begin());
     if (!populateDependencyMatrix(DependencyMatrix, LoopNestDepth,
-                                  OuterMostLoop, LoopList, LI, DI, SE, ORE)) {
+                                  OuterMostLoop, DI, SE, ORE)) {
       LLVM_DEBUG(dbgs() << "Populating dependency matrix failed\n");
       return false;
     }
@@ -1517,11 +1509,11 @@ static bool areOuterLoopExitPHIsSupported(Loop *OuterLoop, Loop *InnerLoop) {
         continue;
 
       // The incoming value is defined in the outer loop latch. Currently we
-      // only support that in case the outer loop latch has a single
-      // predecessor. This guarantees that the outer loop latch is executed if
-      // and only if the inner loop is executed (because tightlyNested()
-      // guarantees that the outer loop header only branches to the inner loop
-      // or the outer loop latch).
+      // only support that in case the outer loop latch has a single predecessor.
+      // This guarantees that the outer loop latch is executed if and only if
+      // the inner loop is executed (because tightlyNested() guarantees that the
+      // outer loop header only branches to the inner loop or the outer loop
+      // latch).
       // FIXME: We could weaken this logic and allow multiple predecessors,
       //        if the values are produced outside the loop latch. We would need
       //        additional logic to update the PHI nodes in the exit block as
diff --git a/llvm/test/Transforms/LoopInterchange/partially-perfect-loop.ll b/llvm/test/Transforms/LoopInterchange/partially-perfect-loop.ll
index 89e63a3f1edaf..43cdcce5f884d 100644
--- a/llvm/test/Transforms/LoopInterchange/partially-perfect-loop.ll
+++ b/llvm/test/Transforms/LoopInterchange/partially-perfect-loop.ll
@@ -17,25 +17,6 @@ define void @f(ptr noalias %A, ptr noalias %B) {
 ; CHECK-NEXT:  [[FOR_J_HEADER:.*]]:
 ; CHECK-NEXT:    br label %[[FOR_R_HEADER_SPLIT1:.*]]
 ; CHECK:       [[FOR_R_HEADER_SPLIT1]]:
-<<<<<<< HEAD
-; CHECK-NEXT:    [[I:%.*]] = phi i64 [ 0, %[[FOR_J_HEADER]] ], [ [[I_NEXT:%.*]], %[[FOR_I_LATCH:.*]] ]
-; CHECK-NEXT:    br label %[[FOR_R_HEADER:.*]]
-; CHECK:       [[FOR_R_HEADER]]:
-; CHECK-NEXT:    [[J:%.*]] = phi i64 [ 0, %[[FOR_R_HEADER_SPLIT1]] ], [ [[J_NEXT:%.*]], %[[FOR_J_LATCH:.*]] ]
-; CHECK-NEXT:    br label %[[FOR_J_HEADER_PREHEADER:.*]]
-; CHECK:       [[FOR_J_HEADER_PREHEADER]]:
-; CHECK-NEXT:    [[R:%.*]] = phi i64 [ 0, %[[FOR_R_HEADER]] ], [ [[TMP0:%.*]], %[[FOR_J_HEADER_PREHEADER]] ]
-; CHECK-NEXT:    [[A_ELEMENT:%.*]] = getelementptr [64 x i8], ptr [[A]], i64 [[J]], i64 [[R]]
-; CHECK-NEXT:    store i8 0, ptr [[A_ELEMENT]], align 1
-; CHECK-NEXT:    [[TMP0]] = add i64 [[R]], 1
-; CHECK-NEXT:    [[TMP1:%.*]] = icmp eq i64 [[TMP0]], 64
-; CHECK-NEXT:    br i1 [[TMP1]], label %[[FOR_J_LATCH]], label %[[FOR_J_HEADER_PREHEADER]]
-; CHECK:       [[FOR_J_LATCH]]:
-; CHECK-NEXT:    [[J_NEXT]] = add i64 [[J]], 1
-; CHECK-NEXT:    [[J_DONE:%.*]] = icmp eq i64 [[J_NEXT]], 64
-; CHECK-NEXT:    br i1 [[J_DONE]], label %[[FOR_L_HEADER_PREHEADER:.*]], label %[[FOR_R_HEADER]]
-; CHECK:       [[FOR_L_HEADER_PREHEADER]]:
-=======
 ; CHECK-NEXT:    [[I:%.*]] = phi i64 [ 0, %[[FOR_J_HEADER]] ], [ [[I_NEXT:%.*]], %[[FOR_I_LATCH1:.*]] ]
 ; CHECK-NEXT:    br label %[[FOR_R_HEADER:.*]]
 ; CHECK:       [[FOR_J_HEADER_PREHEADER1:.*]]:
@@ -63,20 +44,10 @@ define void @f(ptr noalias %A, ptr noalias %B) {
 ; CHECK-NEXT:    [[J_DONE:%.*]] = icmp eq i64 [[J_NEXT]], 64
 ; CHECK-NEXT:    br i1 [[J_DONE]], label %[[FOR_J_LATCH]], label %[[FOR_J_HEADER_PREHEADER]]
 ; CHECK:       [[FOR_L_HEADER_PREHEADER:.*]]:
->>>>>>> effa84a158f2 ([LoopInterchange] Updated the tests for imperfect-loop-nests)
 ; CHECK-NEXT:    br label %[[FOR_L_HEADER:.*]]
 ; CHECK:       [[FOR_L_HEADER]]:
 ; CHECK-NEXT:    [[K:%.*]] = phi i64 [ [[K_NEXT:%.*]], %[[FOR_K_LATCH:.*]] ], [ 0, %[[FOR_L_HEADER_PREHEADER]] ]
 ; CHECK-NEXT:    br label %[[FOR_K_HEADER_PREHEADER:.*]]
-<<<<<<< HEAD
-; CHECK:       [[FOR_K_HEADER_PREHEADER]]:
-; CHECK-NEXT:    [[L:%.*]] = phi i64 [ 0, %[[FOR_L_HEADER]] ], [ [[TMP2:%.*]], %[[FOR_K_HEADER_PREHEADER]] ]
-; CHECK-NEXT:    [[B_ELEMENT:%.*]] = getelementptr [64 x i8], ptr [[B]], i64 [[L]], i64 [[K]]
-; CHECK-NEXT:    store i8 0, ptr [[B_ELEMENT]], align 1
-; CHECK-NEXT:    [[TMP2]] = add i64 [[L]], 1
-; CHECK-NEXT:    [[TMP3:%.*]] = icmp eq i64 [[TMP2]], 64
-; CHECK-NEXT:    br i1 [[TMP3]], label %[[FOR_K_LATCH]], label %[[FOR_K_HEADER_PREHEADER]]
-=======
 ; CHECK:       [[FOR_L_HEADER_PREHEADER1]]:
 ; CHECK-NEXT:    br label %[[FOR_L_HEADER1:.*]]
 ; CHECK:       [[FOR_L_HEADER1]]:
@@ -92,16 +63,11 @@ define void @f(ptr noalias %A, ptr noalias %B) {
 ; CHECK-NEXT:    [[TMP6]] = add i64 [[L]], 1
 ; CHECK-NEXT:    [[TMP7:%.*]] = icmp eq i64 [[TMP6]], 64
 ; CHECK-NEXT:    br i1 [[TMP7]], label %[[FOR_I_LATCH1]], label %[[FOR_L_HEADER1]]
->>>>>>> effa84a158f2 ([LoopInterchange] Updated the tests for imperfect-loop-nests)
 ; CHECK:       [[FOR_K_LATCH]]:
 ; CHECK-NEXT:    [[K_NEXT]] = add i64 [[K]], 1
 ; CHECK-NEXT:    [[K_DONE:%.*]] = icmp eq i64 [[K_NEXT]], 64
 ; CHECK-NEXT:    br i1 [[K_DONE]], label %[[FOR_I_LATCH]], label %[[FOR_L_HEADER]]
-<<<<<<< HEAD
-; CHECK:       [[FOR_I_LATCH]]:
-=======
 ; CHECK:       [[FOR_I_LATCH1]]:
->>>>>>> effa84a158f2 ([LoopInterchange] Updated the tests for imperfect-loop-nests)
 ; CHECK-NEXT:    [[I_NEXT]] = add i64 [[I]], 1
 ; CHECK-NEXT:    [[I_DONE:%.*]] = icmp eq i64 [[I_NEXT]], 64
 ; CHECK-NEXT:    br i1 [[I_DONE]], label %[[EXIT:.*]], label %[[FOR_R_HEADER_SPLIT1]]

>From a5ddcb4cb7c95e0db1dfffb05a22e225b2623c35 Mon Sep 17 00:00:00 2001
From: rohgarg <rohgarg at qti.qualcomm.com>
Date: Tue, 16 Jun 2026 08:35:54 -0700
Subject: [PATCH 5/9] [LoopInterchange] Remove XFAIL and fix issues causing
 failures

---
 .../lib/Transforms/Scalar/LoopInterchange.cpp |  4 ++
 .../LoopInterchange/bail-out-one-loop.ll      |  3 +-
 .../LoopInterchange/large-nested-6d.ll        | 42 ++++++++++++-------
 3 files changed, 32 insertions(+), 17 deletions(-)

diff --git a/llvm/lib/Transforms/Scalar/LoopInterchange.cpp b/llvm/lib/Transforms/Scalar/LoopInterchange.cpp
index e76a20ee6beba..2d957e1a7cbbb 100644
--- a/llvm/lib/Transforms/Scalar/LoopInterchange.cpp
+++ b/llvm/lib/Transforms/Scalar/LoopInterchange.cpp
@@ -721,6 +721,10 @@ struct LoopInterchange {
 
   bool run(LoopNest &LN) {
     SmallVector<SmallVector<Loop *, 8>, 4> LoopLists = collectPerfectNests(LN);
+    if(LoopLists.empty()){
+      LLVM_DEBUG(dbgs() << "No Valid candidates for loop interchange.\n");
+      return false;
+    }
     bool Changed = false;
     for (SmallVector<Loop *, 8> &LoopList : LoopLists) {
       // Ensure minimum depth of the loop nest to do the interchange.
diff --git a/llvm/test/Transforms/LoopInterchange/bail-out-one-loop.ll b/llvm/test/Transforms/LoopInterchange/bail-out-one-loop.ll
index efe25f8455655..d2d5889d2f964 100644
--- a/llvm/test/Transforms/LoopInterchange/bail-out-one-loop.ll
+++ b/llvm/test/Transforms/LoopInterchange/bail-out-one-loop.ll
@@ -1,7 +1,6 @@
 ; REQUIRES: asserts
 
 ; RUN: opt < %s -passes=loop-interchange -debug -disable-output 2>&1 | FileCheck %s
-; XFAIL: *
 
 target datalayout = "e-m:e-p270:32:32-p271:32:32-p272:64:64-i8:8:32-i16:16:32-i64:64-i128:128-n32:64-S128-Fn32"
 
@@ -16,7 +15,7 @@ target datalayout = "e-m:e-p270:32:32-p271:32:32-p272:64:64-i8:8:32-i16:16:32-i6
 ; CHECK-NOT: Delinearizing
 ; CHECK-NOT: Strides:
 ; CHECK-NOT: Terms:
-; CHECK: Unsupported depth of loop nest 1, the supported range is [2, 10].
+; CHECK: No Valid candidates for loop interchange.
 
 define void @foo() {
 entry:
diff --git a/llvm/test/Transforms/LoopInterchange/large-nested-6d.ll b/llvm/test/Transforms/LoopInterchange/large-nested-6d.ll
index dfe4857b52485..f2c4276f602de 100644
--- a/llvm/test/Transforms/LoopInterchange/large-nested-6d.ll
+++ b/llvm/test/Transforms/LoopInterchange/large-nested-6d.ll
@@ -1,6 +1,5 @@
 ; RUN: opt < %s -passes=loop-interchange -cache-line-size=64 -pass-remarks='loop-interchange' -pass-remarks-missed='loop-interchange' -pass-remarks-output=%t -disable-output -S
 ; RUN: FileCheck --input-file=%t %s
-; XFAIL: *
 
 target datalayout = "e-m:e-p270:32:32-p271:32:32-p272:64:64-i8:8:32-i16:16:32-i64:64-i128:128-n32:64-S128-Fn32"
 
@@ -56,16 +55,33 @@ target datalayout = "e-m:e-p270:32:32-p271:32:32-p272:64:64-i8:8:32-i16:16:32-i6
 ;      Dst:  store double %46, ptr %48, align 8
 ;
 ;
-; CHECK:       --- !Missed
+; CHECK:       --- !Analysis
 ; CHECK-NEXT:  Pass:            loop-interchange
-; CHECK-NEXT:  Name:            UnsupportedLoopNestDepth
+; CHECK-NEXT:  Name:            Dependence
+; CHECK-NEXT:  Function:        test
+; CHECK-NEXT:  Args:
+; CHECK-NEXT:    - String:          Computed dependence info, invoking the transform.
+; CHECK-NEXT:  ...
+; CHECK-NEXT:  --- !Missed
+; CHECK-NEXT:  Pass:            loop-interchange
+; CHECK-NEXT:  Name:            Dependence
 ; CHECK-NEXT:  Function:        test
 ; CHECK-NEXT:  Args:
-; CHECK-NEXT:    - String:          'Unsupported depth of loop nest, the supported range is ['
-; CHECK-NEXT:    - String:          '2'
-; CHECK-NEXT:    - String:          ', '
-; CHECK-NEXT:    - String:          '10'
-; CHECK-NEXT:    - String:          "].\n"
+; CHECK-NEXT:    - String:          Cannot interchange loops due to dependences.
+; CHECK-NEXT:  ...
+; CHECK-NEXT:  --- !Missed
+; CHECK-NEXT:  Pass:            loop-interchange
+; CHECK-NEXT:  Name:            Dependence
+; CHECK-NEXT:  Function:        test
+; CHECK-NEXT:  Args:
+; CHECK-NEXT:    - String:          Cannot interchange loops due to dependences.
+; CHECK-NEXT:  ...
+; CHECK-NEXT:  --- !Missed
+; CHECK-NEXT:  Pass:            loop-interchange
+; CHECK-NEXT:  Name:            Dependence
+; CHECK-NEXT:  Function:        test
+; CHECK-NEXT:  Args:
+; CHECK-NEXT:    - String:          All loops have dependencies in all directions.
 ; CHECK-NEXT:  ...
 ; CHECK-NEXT:  --- !Analysis
 ; CHECK-NEXT:  Pass:            loop-interchange
@@ -81,16 +97,12 @@ target datalayout = "e-m:e-p270:32:32-p271:32:32-p272:64:64-i8:8:32-i16:16:32-i6
 ; CHECK-NEXT:  Args:
 ; CHECK-NEXT:    - String:          Cannot interchange loops due to dependences.
 ; CHECK-NEXT:  ...
-; CHECK-NEXT:  --- !Missed
+; CHECK-NEXT:  --- !Analysis
 ; CHECK-NEXT:  Pass:            loop-interchange
-; CHECK-NEXT:  Name:            UnsupportedLoopNestDepth
+; CHECK-NEXT:  Name:            Dependence
 ; CHECK-NEXT:  Function:        test
 ; CHECK-NEXT:  Args:
-; CHECK-NEXT:    - String:          'Unsupported depth of loop nest, the supported range is ['
-; CHECK-NEXT:    - String:          '2'
-; CHECK-NEXT:    - String:          ', '
-; CHECK-NEXT:    - String:          '10'
-; CHECK-NEXT:    - String:          "].\n"
+; CHECK-NEXT:    - String:          Computed dependence info, invoking the transform.
 ; CHECK-NEXT:  ...
 ; CHECK-NEXT:  --- !Analysis
 ; CHECK-NEXT:  Pass:            loop-interchange

>From f879c4dca956893879d7f9b0cc7b4b4cde578ded Mon Sep 17 00:00:00 2001
From: rohgarg <rohgarg at qti.qualcomm.com>
Date: Wed, 17 Jun 2026 00:10:29 -0700
Subject: [PATCH 6/9] Adding tests for padding condition

---
 .../dependency-matrix-padding.ll              | 118 ++++++++++++++++++
 1 file changed, 118 insertions(+)
 create mode 100644 llvm/test/Transforms/LoopInterchange/dependency-matrix-padding.ll

diff --git a/llvm/test/Transforms/LoopInterchange/dependency-matrix-padding.ll b/llvm/test/Transforms/LoopInterchange/dependency-matrix-padding.ll
new file mode 100644
index 0000000000000..00ea733be36da
--- /dev/null
+++ b/llvm/test/Transforms/LoopInterchange/dependency-matrix-padding.ll
@@ -0,0 +1,118 @@
+; RUN: opt < %s -passes=loop-interchange -loop-interchange-profitabilities=ignore -debug-only=loop-interchange,da -disable-output -S 2>%t
+; RUN: FileCheck --input-file=%t %s
+; Generated by AI.
+;
+; This test focuses exclusively on validating the padding logic in the dependency
+; matrix construction and ensuring that matrix slicing preserves proper alignment 
+; with the corresponding loops.
+;
+; Corresponding C code:
+;
+;   for (int i = 0; i < 32; ++i) {
+;     for (int j = 0; j < 32; ++j) {
+;       int sum = 0;
+;       for (int k = 0; k < 32; ++k)
+;         sum += k;
+;       S[j] = sum + S[j-1]; 
+;     }
+;     for (int x = 0; x < 32; ++x)
+;       D[x] = 0;
+;   }
+;
+; CHECK: Processing LoopList of size = 2 containing the following loops:
+; CHECK-NEXT:   - Loop at depth 2 containing: %loop.j.header<header>,%loop.k.header,%loop.k.latch,%loop.j.latch<latch><exiting>
+; CHECK-NEXT:     Loop at depth 3 containing: %loop.k.header<header>,%loop.k.latch<latch><exiting>
+; CHECK-NEXT:   - Loop at depth 3 containing: %loop.k.header<header>,%loop.k.latch<latch><exiting>
+; CHECK: Found 2 Loads and Stores to analyze
+; CHECK: common nesting levels = 2
+; CHECK: loops = {2}
+;
+; DA reports the subscript only lives at loop depth 2 (loop.j), confirming
+; that loop.i contributes '=' and loop.j contributes the non-trivial '>'.
+; DA reports the anti-dependence with distance -1 and direction '<' before
+; normalization.  normalize() flips the negative distance, turning '<' into
+; '>'.
+; CHECK: Result = anti [S -1|<]!
+;
+; Distance 0 at loop.j level gives '=' for both levels (loop.i, loop.j).
+; CHECK: common nesting levels = 2
+; CHECK: loops = {2}
+; CHECK: Result = output [S 0]!
+;
+; For both dependencies, Dep.size() = 2  after DA fill and before padding
+;   L->getLoopDepth() = 2  (loop.j is the outermost loop of the subnest)
+;   Level             = 2  (subnest [loop.j, loop.k] has 2 loops)
+;   L->getLoopDepth() + Level - 1 = 2 + 2 - 1 = 3
+;   2 < 3  =>  padding fires once, appending 'I'
+;
+;   Dep 1 after padding:  ['=', '>', 'I']  (size 3)
+;   Dep 2 after padding:  ['=', '=', 'I']  (size 3)
+;
+; The first column of this matrix should be dropped.
+; CHECK: Dependency matrix before interchange:
+; CHECK-NEXT: > I
+; CHECK-NEXT: = I
+;
+; CHECK: Failed interchange InnerLoopId = 1 and OuterLoopId = 0 due to dependence
+
+define void @test_padding_nontrivial_direction(ptr noalias %S, ptr noalias %D) {
+entry:
+  br label %loop.i.header
+
+loop.i.header:
+  %i = phi i64 [ 0, %entry ], [ %i.next, %loop.i.latch ]
+  br label %loop.j.header
+
+loop.j.header:
+  %j     = phi i64 [ 0, %loop.i.header ], [ %j.next, %loop.j.latch ]
+  %s.ptr = getelementptr i32, ptr %S, i64 %j
+  br label %loop.k.header
+
+loop.k.header:
+  %k        = phi i64 [ 0, %loop.j.header ], [ %k.next, %loop.k.latch ]
+  %sum      = phi i32 [ 0, %loop.j.header ], [ %sum.next, %loop.k.latch ]
+  %k.trunc  = trunc i64 %k to i32
+  %sum.next = add nsw i32 %sum, %k.trunc
+  br label %loop.k.latch
+
+loop.k.latch:
+  %k.next = add nuw nsw i64 %k, 1
+  %k.done = icmp eq i64 %k.next, 32
+  br i1 %k.done, label %loop.j.latch, label %loop.k.header
+
+loop.j.latch:
+  %sum.lcssa = phi i32 [ %sum.next, %loop.k.latch ]
+  ; Load S[j-1] — reads what the previous j iteration stored into S[j-1].
+  ; This load appears before the store in the same BB, so MemInstr order is
+  ; [load, store].  DA tests (load, store) and finds an anti-dependence with
+  ; distance -1 at the loop.j level; normalize() flips it to direction '>'.
+  ; After padding: ['>', 'I'].  The '>' makes interchange illegal.
+  %jm1    = add i64 %j, -1
+  %s.prev = getelementptr i32, ptr %S, i64 %jm1
+  %s.load = load i32, ptr %s.prev, align 4
+  ; Store S[j] = sum + S[j-1].
+  %val    = add nsw i32 %sum.lcssa, %s.load
+  store i32 %val, ptr %s.ptr, align 4
+  %j.next = add nuw nsw i64 %j, 1
+  %j.done = icmp eq i64 %j.next, 32
+  br i1 %j.done, label %loop.x.header, label %loop.j.header
+
+loop.x.header:
+  %x = phi i64 [ 0, %loop.j.latch ], [ %x.next, %loop.x.latch ]
+  %d.ptr = getelementptr i8, ptr %D, i64 %x
+  store i8 0, ptr %d.ptr, align 1
+  br label %loop.x.latch
+
+loop.x.latch:
+  %x.next = add nuw nsw i64 %x, 1
+  %x.done = icmp eq i64 %x.next, 32
+  br i1 %x.done, label %loop.i.latch, label %loop.x.header
+
+loop.i.latch:
+  %i.next = add nuw nsw i64 %i, 1
+  %i.done = icmp eq i64 %i.next, 32
+  br i1 %i.done, label %exit, label %loop.i.header
+
+exit:
+  ret void
+}

>From 5fa484c9708bfefdde39cdc66ca89b3ecc7eeb06 Mon Sep 17 00:00:00 2001
From: rohgarg <rohgarg at qti.qualcomm.com>
Date: Wed, 17 Jun 2026 03:11:16 -0700
Subject: [PATCH 7/9] [LoopInterchange] Added test for early bailout when loop
 list is empty.

---
 .../no-partially-perfect-subnest.ll           | 62 +++++++++++++++++++
 1 file changed, 62 insertions(+)
 create mode 100644 llvm/test/Transforms/LoopInterchange/no-partially-perfect-subnest.ll

diff --git a/llvm/test/Transforms/LoopInterchange/no-partially-perfect-subnest.ll b/llvm/test/Transforms/LoopInterchange/no-partially-perfect-subnest.ll
new file mode 100644
index 0000000000000..e94d4a22a9125
--- /dev/null
+++ b/llvm/test/Transforms/LoopInterchange/no-partially-perfect-subnest.ll
@@ -0,0 +1,62 @@
+; RUN: opt < %s -passes=loop-interchange -loop-interchange-profitabilities=ignore -debug-only=loop-interchange -disable-output -S 2>%t
+; RUN: FileCheck --input-file=%t %s
+
+; There is no partially-perfect subnest here. Every innermost loop has a parent
+; with multiple child loops, so collectPerfectNests() should return an empty
+; list and loop-interchange should bail out immediately without attempting or
+; performing any interchange.
+;
+; Corresponding C code:
+;
+;   for (int i = 0; i < 16; ++i) {
+;     for (int j = 0; j < 16; ++j)
+;       A[i][j] = 0;
+;
+;     for (int k = 0; k < 16; ++k)
+;       B[i][k] = 0;
+;   }
+;
+; CHECK: No Valid candidates for loop interchange.
+
+define void @no_partially_perfect_subnest(ptr noalias %A, ptr noalias %B) {
+entry:
+  br label %loop.i.header
+
+loop.i.header:
+  %i = phi i64 [ 0, %entry ], [ %i.next, %loop.i.latch ]
+  br label %loop.j.header
+
+loop.j.header:
+  %j = phi i64 [ 0, %loop.i.header ], [ %j.next, %loop.j.latch ]
+  %a.row.base = mul nuw nsw i64 %i, 16
+  %a.index = add nuw nsw i64 %a.row.base, %j
+  %a.element.ptr = getelementptr i8, ptr %A, i64 %a.index
+  store i8 1, ptr %a.element.ptr, align 1
+  br label %loop.j.latch
+
+loop.j.latch:
+  %j.next = add nuw nsw i64 %j, 1
+  %j.done = icmp eq i64 %j.next, 16
+  br i1 %j.done, label %loop.k.header, label %loop.j.header
+
+loop.k.header:
+  %k = phi i64 [ 0, %loop.j.latch ], [ %k.next, %loop.k.latch ]
+  %b.row.base = mul nuw nsw i64 %i, 16
+  %b.index = add nuw nsw i64 %b.row.base, %k
+  %b.element.ptr = getelementptr i8, ptr %B, i64 %b.index
+  store i8 2, ptr %b.element.ptr, align 1
+  br label %loop.k.latch
+
+loop.k.latch:
+  %k.next = add nuw nsw i64 %k, 1
+  %k.done = icmp eq i64 %k.next, 16
+  br i1 %k.done, label %loop.i.latch, label %loop.k.header
+
+loop.i.latch:
+  %i.next = add nuw nsw i64 %i, 1
+  %i.done = icmp eq i64 %i.next, 16
+  br i1 %i.done, label %exit, label %loop.i.header
+
+exit:
+  ret void
+}

>From aeb90f67f7f7585f71a98617f14cba32528dce10 Mon Sep 17 00:00:00 2001
From: rohgarg <rohgarg at qti.qualcomm.com>
Date: Thu, 18 Jun 2026 00:53:03 -0700
Subject: [PATCH 8/9] [LoopInterchange] Added test for missed Loop pair for
 interchange and ran clang-format

---
 .../lib/Transforms/Scalar/LoopInterchange.cpp |   2 +-
 .../missed-outer-prefix-subnest.ll            | 125 ++++++++++++++++++
 2 files changed, 126 insertions(+), 1 deletion(-)
 create mode 100644 llvm/test/Transforms/LoopInterchange/missed-outer-prefix-subnest.ll

diff --git a/llvm/lib/Transforms/Scalar/LoopInterchange.cpp b/llvm/lib/Transforms/Scalar/LoopInterchange.cpp
index 2d957e1a7cbbb..0c1f811844963 100644
--- a/llvm/lib/Transforms/Scalar/LoopInterchange.cpp
+++ b/llvm/lib/Transforms/Scalar/LoopInterchange.cpp
@@ -721,7 +721,7 @@ struct LoopInterchange {
 
   bool run(LoopNest &LN) {
     SmallVector<SmallVector<Loop *, 8>, 4> LoopLists = collectPerfectNests(LN);
-    if(LoopLists.empty()){
+    if (LoopLists.empty()) {
       LLVM_DEBUG(dbgs() << "No Valid candidates for loop interchange.\n");
       return false;
     }
diff --git a/llvm/test/Transforms/LoopInterchange/missed-outer-prefix-subnest.ll b/llvm/test/Transforms/LoopInterchange/missed-outer-prefix-subnest.ll
new file mode 100644
index 0000000000000..c04bfb6f203de
--- /dev/null
+++ b/llvm/test/Transforms/LoopInterchange/missed-outer-prefix-subnest.ll
@@ -0,0 +1,125 @@
+; RUN: opt < %s -passes=loop-interchange -loop-interchange-profitabilities=ignore -debug-only=loop-interchange -disable-output -S 2>%t
+; RUN: FileCheck --input-file=%t %s
+
+; This test documents a currently missed optimization opportunity.
+;
+; collectPerfectNests() walks up from each innermost loop and stops as soon as
+; it reaches a loop with more than one subloop. But because [loop.i, loop.j] is never put in
+; any LoopList, the pass never analyses or attempts that interchange — the
+; opportunity is silently missed.
+;
+; Corresponding C code:
+;
+;   for (int i = 0; i < 8; ++i)
+;     for (int j = 0; j < 8; ++j) {
+;       A[i][j] = 0;                // missed: i/j interchange
+;       for (int k = 0; k < 8; ++k) {
+;         for (int l = 0; l < 8; ++l)
+;           for (int m = 0; m < 8; ++m)
+;             Left[m][l] = 0;       // interchanged: l/m swapped
+;         for (int n = 0; n < 8; ++n)
+;           for (int o = 0; o < 8; ++o)
+;             Right[o][n] = 0;      // interchanged: n/o swapped
+;       }
+;     }
+;
+;
+; CHECK:      Processing LoopList of size = 2 containing the following loops:
+; CHECK-NEXT:   - Loop at depth 4 containing: %loop.l.header<header>,%loop.m.header,%loop.m.latch,%loop.l.latch<latch><exiting>
+; CHECK-NEXT:     Loop at depth 5 containing: %loop.m.header<header>,%loop.m.latch<latch><exiting>
+; CHECK-NEXT:   - Loop at depth 5 containing: %loop.m.header<header>,%loop.m.latch<latch><exiting>
+; CHECK: Loops interchanged: outer loop 'loop.l.header' and inner loop 'loop.m.header'
+;
+; CHECK:      Processing LoopList of size = 2 containing the following loops:
+; CHECK-NEXT:   - Loop at depth 4 containing: %loop.n.header<header>,%loop.o.header,%loop.o.latch,%loop.n.latch<latch><exiting>
+; CHECK-NEXT:     Loop at depth 5 containing: %loop.o.header<header>,%loop.o.latch<latch><exiting>
+; CHECK-NEXT:   - Loop at depth 5 containing: %loop.o.header<header>,%loop.o.latch<latch><exiting>
+; CHECK: Loops interchanged: outer loop 'loop.n.header' and inner loop 'loop.o.header'
+;
+;
+; CHECK-NOT: loop.i.header
+; CHECK-NOT: loop.j.header
+
+define void @missed_outer_prefix_subnest(ptr noalias %Left, ptr noalias %Right,
+                                         ptr noalias %A) {
+entry:
+  br label %loop.i.header
+
+loop.i.header:
+  %i = phi i64 [ 0, %entry ], [ %i.next, %loop.i.latch ]
+  br label %loop.j.header
+
+loop.j.header:
+  %j = phi i64 [ 0, %loop.i.header ], [ %j.next, %loop.j.latch ]
+  %a.row = mul nuw nsw i64 %i, 8
+  %a.idx = add nuw nsw i64 %a.row, %j
+  %a.ptr = getelementptr i8, ptr %A, i64 %a.idx
+  store i8 0, ptr %a.ptr, align 1
+  br label %loop.k.header
+
+loop.k.header:
+  %k = phi i64 [ 0, %loop.j.header ], [ %k.next, %loop.k.latch ]
+  br label %loop.l.header
+
+loop.l.header:
+  %l = phi i64 [ 0, %loop.k.header ], [ %l.next, %loop.l.latch ]
+  br label %loop.m.header
+
+loop.m.header:
+  %m = phi i64 [ 0, %loop.l.header ], [ %m.next, %loop.m.latch ]
+  %left.row.base = mul nuw nsw i64 %m, 8
+  %left.index = add nuw nsw i64 %left.row.base, %l
+  %left.element.ptr = getelementptr i8, ptr %Left, i64 %left.index
+  store i8 0, ptr %left.element.ptr, align 1
+  br label %loop.m.latch
+
+loop.m.latch:
+  %m.next = add nuw nsw i64 %m, 1
+  %m.done = icmp eq i64 %m.next, 8
+  br i1 %m.done, label %loop.l.latch, label %loop.m.header
+
+loop.l.latch:
+  %l.next = add nuw nsw i64 %l, 1
+  %l.done = icmp eq i64 %l.next, 8
+  br i1 %l.done, label %loop.n.header, label %loop.l.header
+
+loop.n.header:
+  %n = phi i64 [ 0, %loop.l.latch ], [ %n.next, %loop.n.latch ]
+  br label %loop.o.header
+
+loop.o.header:
+  %o = phi i64 [ 0, %loop.n.header ], [ %o.next, %loop.o.latch ]
+  %right.row.base = mul nuw nsw i64 %o, 8
+  %right.index = add nuw nsw i64 %right.row.base, %n
+  %right.element.ptr = getelementptr i8, ptr %Right, i64 %right.index
+  store i8 0, ptr %right.element.ptr, align 1
+  br label %loop.o.latch
+
+loop.o.latch:
+  %o.next = add nuw nsw i64 %o, 1
+  %o.done = icmp eq i64 %o.next, 8
+  br i1 %o.done, label %loop.n.latch, label %loop.o.header
+
+loop.n.latch:
+  %n.next = add nuw nsw i64 %n, 1
+  %n.done = icmp eq i64 %n.next, 8
+  br i1 %n.done, label %loop.k.latch, label %loop.n.header
+
+loop.k.latch:
+  %k.next = add nuw nsw i64 %k, 1
+  %k.done = icmp eq i64 %k.next, 8
+  br i1 %k.done, label %loop.j.latch, label %loop.k.header
+
+loop.j.latch:
+  %j.next = add nuw nsw i64 %j, 1
+  %j.done = icmp eq i64 %j.next, 8
+  br i1 %j.done, label %loop.i.latch, label %loop.j.header
+
+loop.i.latch:
+  %i.next = add nuw nsw i64 %i, 1
+  %i.done = icmp eq i64 %i.next, 8
+  br i1 %i.done, label %exit, label %loop.i.header
+
+exit:
+  ret void
+}

>From 8afe1e47b15745e41ec20381c7e6873031a9e4ab Mon Sep 17 00:00:00 2001
From: rohgarg <rohgarg at qti.qualcomm.com>
Date: Sun, 19 Jul 2026 23:33:57 -0700
Subject: [PATCH 9/9] [LoopInterchange] Add more partially-perfect loop tests

---
 .../dependency-matrix-padding.ll              |  41 +++---
 .../LoopInterchange/partially-perfect-loop.ll | 124 ++++++++++++++++++
 2 files changed, 141 insertions(+), 24 deletions(-)

diff --git a/llvm/test/Transforms/LoopInterchange/dependency-matrix-padding.ll b/llvm/test/Transforms/LoopInterchange/dependency-matrix-padding.ll
index 00ea733be36da..9f790be2c4a65 100644
--- a/llvm/test/Transforms/LoopInterchange/dependency-matrix-padding.ll
+++ b/llvm/test/Transforms/LoopInterchange/dependency-matrix-padding.ll
@@ -1,6 +1,5 @@
-; RUN: opt < %s -passes=loop-interchange -loop-interchange-profitabilities=ignore -debug-only=loop-interchange,da -disable-output -S 2>%t
-; RUN: FileCheck --input-file=%t %s
-; Generated by AI.
+; RUN: opt < %s -passes=loop-interchange -loop-interchange-profitabilities=ignore -debug-only=loop-interchange,da -disable-output 2>&1 | FileCheck %s
+; REQUIRES: asserts
 ;
 ; This test focuses exclusively on validating the padding logic in the dependency
 ; matrix construction and ensuring that matrix slicing preserves proper alignment 
@@ -19,26 +18,11 @@
 ;       D[x] = 0;
 ;   }
 ;
-; CHECK: Processing LoopList of size = 2 containing the following loops:
-; CHECK-NEXT:   - Loop at depth 2 containing: %loop.j.header<header>,%loop.k.header,%loop.k.latch,%loop.j.latch<latch><exiting>
-; CHECK-NEXT:     Loop at depth 3 containing: %loop.k.header<header>,%loop.k.latch<latch><exiting>
-; CHECK-NEXT:   - Loop at depth 3 containing: %loop.k.header<header>,%loop.k.latch<latch><exiting>
-; CHECK: Found 2 Loads and Stores to analyze
-; CHECK: common nesting levels = 2
-; CHECK: loops = {2}
-;
-; DA reports the subscript only lives at loop depth 2 (loop.j), confirming
+; For the first dependency, DA reports the subscript only lives at loop depth 2 (loop.j); confirming
 ; that loop.i contributes '=' and loop.j contributes the non-trivial '>'.
-; DA reports the anti-dependence with distance -1 and direction '<' before
-; normalization.  normalize() flips the negative distance, turning '<' into
-; '>'.
-; CHECK: Result = anti [S -1|<]!
-;
-; Distance 0 at loop.j level gives '=' for both levels (loop.i, loop.j).
-; CHECK: common nesting levels = 2
-; CHECK: loops = {2}
-; CHECK: Result = output [S 0]!
 ;
+; For the second dependency, Distance 0 at loop.j level gives '=' for both levels (loop.i, loop.j).
+; 
 ; For both dependencies, Dep.size() = 2  after DA fill and before padding
 ;   L->getLoopDepth() = 2  (loop.j is the outermost loop of the subnest)
 ;   Level             = 2  (subnest [loop.j, loop.k] has 2 loops)
@@ -48,13 +32,22 @@
 ;   Dep 1 after padding:  ['=', '>', 'I']  (size 3)
 ;   Dep 2 after padding:  ['=', '=', 'I']  (size 3)
 ;
-; The first column of this matrix should be dropped.
+;   The first column of this matrix should be dropped. And the Final Dependecy Matrix before Interchange should be:
+;   Dep 1: ['>', 'I'] (Size 2)
+;   Dep 2: ['=', 'I'] (Size 3)
+;
+; CHECK: Found 2 Loads and Stores to analyze
+; CHECK: common nesting levels = 2
+; CHECK: loops = {2}
+; CHECK: Result = anti [S -1|<]!
+; CHECK: common nesting levels = 2
+; CHECK: loops = {2}
+; CHECK: Result = output [S 0]!
 ; CHECK: Dependency matrix before interchange:
 ; CHECK-NEXT: > I
 ; CHECK-NEXT: = I
-;
 ; CHECK: Failed interchange InnerLoopId = 1 and OuterLoopId = 0 due to dependence
-
+;
 define void @test_padding_nontrivial_direction(ptr noalias %S, ptr noalias %D) {
 entry:
   br label %loop.i.header
diff --git a/llvm/test/Transforms/LoopInterchange/partially-perfect-loop.ll b/llvm/test/Transforms/LoopInterchange/partially-perfect-loop.ll
index 43cdcce5f884d..a04d1dd4229f9 100644
--- a/llvm/test/Transforms/LoopInterchange/partially-perfect-loop.ll
+++ b/llvm/test/Transforms/LoopInterchange/partially-perfect-loop.ll
@@ -1,5 +1,8 @@
 ; NOTE: Assertions have been autogenerated by utils/update_test_checks.py UTC_ARGS: --version 6
+; REQUIRES: asserts
 ; RUN: opt -S -passes='loop-interchange' -loop-interchange-profitabilities=ignore < %s | FileCheck %s
+; RUN: opt < %s -passes=loop-interchange -loop-interchange-profitabilities=ignore -debug-only=loop-interchange -disable-output -S 2>%t
+; RUN: FileCheck --input-file=%t %s --check-prefix=MISS-PREFIX
 ;
 ; for (i = 0; i < 64; i++) {
 ;   for (j = 0; j < 64; j++)
@@ -123,3 +126,124 @@ for.i.latch:
 exit:
   ret void
 }
+
+; This case shows a current limitation.
+;
+; collectPerfectNests() walks up from each innermost loop and stops as soon as
+; it reaches a loop with more than one subloop. But because [loop.i, loop.j] is never put in
+; any LoopList, the pass never analyses or attempts that interchange — the
+; opportunity is silently missed.
+;
+; Corresponding C code:
+;
+;   for (int i = 0; i < 8; ++i)
+;     for (int j = 0; j < 8; ++j) {
+;       A[i][j] = 0;                // missed: i/j interchange
+;       for (int k = 0; k < 8; ++k) {
+;         for (int l = 0; l < 8; ++l)
+;           for (int m = 0; m < 8; ++m)
+;             Left[m][l] = 0;       // interchanged: l/m swapped
+;         for (int n = 0; n < 8; ++n)
+;           for (int o = 0; o < 8; ++o)
+;             Right[o][n] = 0;      // interchanged: n/o swapped
+;       }
+;     }
+;
+
+define void @missed_outer_prefix_subnest(ptr noalias %Left, ptr noalias %Right,
+                                         ptr noalias %A) {
+; MISS-PREFIX:      Processing LoopList of size = 2 containing the following loops:
+; MISS-PREFIX:        - Loop at depth 4 containing: %loop.l.header<header>,%loop.m.header,%loop.m.latch,%loop.l.latch<latch><exiting>
+; MISS-PREFIX-NEXT:     Loop at depth 5 containing: %loop.m.header<header>,%loop.m.latch<latch><exiting>
+; MISS-PREFIX-NEXT:   - Loop at depth 5 containing: %loop.m.header<header>,%loop.m.latch<latch><exiting>
+; MISS-PREFIX: Loops interchanged: outer loop 'loop.l.header' and inner loop 'loop.m.header'
+;
+; MISS-PREFIX:      Processing LoopList of size = 2 containing the following loops:
+; MISS-PREFIX:        - Loop at depth 4 containing: %loop.n.header<header>,%loop.o.header,%loop.o.latch,%loop.n.latch<latch><exiting>
+; MISS-PREFIX-NEXT:     Loop at depth 5 containing: %loop.o.header<header>,%loop.o.latch<latch><exiting>
+; MISS-PREFIX-NEXT:   - Loop at depth 5 containing: %loop.o.header<header>,%loop.o.latch<latch><exiting>
+; MISS-PREFIX: Loops interchanged: outer loop 'loop.n.header' and inner loop 'loop.o.header'
+;
+; MISS-PREFIX-NOT: loop.i.header
+; MISS-PREFIX-NOT: loop.j.header
+entry:
+  br label %loop.i.header
+
+loop.i.header:
+  %i = phi i64 [ 0, %entry ], [ %i.next, %loop.i.latch ]
+  br label %loop.j.header
+
+loop.j.header:
+  %j = phi i64 [ 0, %loop.i.header ], [ %j.next, %loop.j.latch ]
+  %a.row = mul nuw nsw i64 %i, 8
+  %a.idx = add nuw nsw i64 %a.row, %j
+  %a.ptr = getelementptr i8, ptr %A, i64 %a.idx
+  store i8 0, ptr %a.ptr, align 1
+  br label %loop.k.header
+
+loop.k.header:
+  %k = phi i64 [ 0, %loop.j.header ], [ %k.next, %loop.k.latch ]
+  br label %loop.l.header
+
+loop.l.header:
+  %l = phi i64 [ 0, %loop.k.header ], [ %l.next, %loop.l.latch ]
+  br label %loop.m.header
+
+loop.m.header:
+  %m = phi i64 [ 0, %loop.l.header ], [ %m.next, %loop.m.latch ]
+  %left.row.base = mul nuw nsw i64 %m, 8
+  %left.index = add nuw nsw i64 %left.row.base, %l
+  %left.element.ptr = getelementptr i8, ptr %Left, i64 %left.index
+  store i8 0, ptr %left.element.ptr, align 1
+  br label %loop.m.latch
+
+loop.m.latch:
+  %m.next = add nuw nsw i64 %m, 1
+  %m.done = icmp eq i64 %m.next, 8
+  br i1 %m.done, label %loop.l.latch, label %loop.m.header
+
+loop.l.latch:
+  %l.next = add nuw nsw i64 %l, 1
+  %l.done = icmp eq i64 %l.next, 8
+  br i1 %l.done, label %loop.n.header, label %loop.l.header
+
+loop.n.header:
+  %n = phi i64 [ 0, %loop.l.latch ], [ %n.next, %loop.n.latch ]
+  br label %loop.o.header
+
+loop.o.header:
+  %o = phi i64 [ 0, %loop.n.header ], [ %o.next, %loop.o.latch ]
+  %right.row.base = mul nuw nsw i64 %o, 8
+  %right.index = add nuw nsw i64 %right.row.base, %n
+  %right.element.ptr = getelementptr i8, ptr %Right, i64 %right.index
+  store i8 0, ptr %right.element.ptr, align 1
+  br label %loop.o.latch
+
+loop.o.latch:
+  %o.next = add nuw nsw i64 %o, 1
+  %o.done = icmp eq i64 %o.next, 8
+  br i1 %o.done, label %loop.n.latch, label %loop.o.header
+
+loop.n.latch:
+  %n.next = add nuw nsw i64 %n, 1
+  %n.done = icmp eq i64 %n.next, 8
+  br i1 %n.done, label %loop.k.latch, label %loop.n.header
+
+loop.k.latch:
+  %k.next = add nuw nsw i64 %k, 1
+  %k.done = icmp eq i64 %k.next, 8
+  br i1 %k.done, label %loop.j.latch, label %loop.k.header
+
+loop.j.latch:
+  %j.next = add nuw nsw i64 %j, 1
+  %j.done = icmp eq i64 %j.next, 8
+  br i1 %j.done, label %loop.i.latch, label %loop.j.header
+
+loop.i.latch:
+  %i.next = add nuw nsw i64 %i, 1
+  %i.done = icmp eq i64 %i.next, 8
+  br i1 %i.done, label %exit, label %loop.i.header
+
+exit:
+  ret void
+}



More information about the llvm-commits mailing list
diff --git a/llvm/lib/Transforms/Scalar/LoopInterchange.cpp b/llvm/lib/Transforms/Scalar/LoopInterchange.cpp
index 01de3ead6ad60..7617c674681ea 100644
--- a/llvm/lib/Transforms/Scalar/LoopInterchange.cpp
+++ b/llvm/lib/Transforms/Scalar/LoopInterchange.cpp
@@ -176,27 +176,25 @@ static bool populateDependencyMatrix(CharMatrix &DepMatrix, unsigned Level,
   ValueVector MemInstr;
   unsigned NumInsts = 0;
 
-  // Collect memory instructions from the BB which belongs to all loops in the LoopList
-  for (Loop *PathLoop : LoopList) {
-    for (BasicBlock *BB : PathLoop->getBlocksVector()) {
-      // In this iteration we need to handle the BB contained directly by Current Loop 
-      // and all other BB of subloops will be handled in its own in next iterations.
-      if (LI->getLoopFor(BB) != PathLoop)
-        continue;
-      // Scan the BB and collect legal loads and stores.
-      for (Instruction &I : *BB) {
-        if (!isa<Instruction>(I))
+  // Collect memory instructions from the BB which belongs to all loops in the
+  // LoopList
+  for (BasicBlock *BB : L->getBlocksVector()) {
+    // In this iteration we need to handle the BB contained directly by Current
+    // Loop and all other BB of subloops will be handled in its own in next
+    // iterations.
+    if (!llvm::is_contained(LoopList, LI->getLoopFor(BB)))
+      continue;
+    // Scan the BB and collect legal loads and stores.
+    for (Instruction &I : *BB) {
+      NumInsts++;
+      if (auto *Ld = dyn_cast<LoadInst>(&I)) {
+        if (!Ld->isSimple())
           return false;
-        NumInsts++;
-        if (auto *Ld = dyn_cast<LoadInst>(&I)) {
-          if (!Ld->isSimple())
-            return false;
-          MemInstr.push_back(&I);
-        } else if (auto *St = dyn_cast<StoreInst>(&I)) {
-          if (!St->isSimple())
-            return false;
-          MemInstr.push_back(&I);
-        }
+        MemInstr.push_back(&I);
+      } else if (auto *St = dyn_cast<StoreInst>(&I)) {
+        if (!St->isSimple())
+          return false;
+        MemInstr.push_back(&I);
       }
     }
   }
@@ -241,8 +239,9 @@ static bool populateDependencyMatrix(CharMatrix &DepMatrix, unsigned Level,
         // make it non-negative.
         if (D->normalize(SE))
           LLVM_DEBUG(dbgs() << "Negative dependence vector normalized.\n");
-        LLVM_DEBUG(StringRef DepType =
-                       D->isFlow() ? "flow" : D->isAnti() ? "anti" : "output";
+        LLVM_DEBUG(StringRef DepType = D->isFlow()   ? "flow"
+                                       : D->isAnti() ? "anti"
+                                                     : "output";
                    dbgs() << "Found " << DepType
                           << " dependency between Src and Dst\n"
                           << " Src:" << *Src << "\n Dst:" << *Dst << '\n');
@@ -274,10 +273,16 @@ static bool populateDependencyMatrix(CharMatrix &DepMatrix, unsigned Level,
           Dep.assign(Level, '*');
         }
 
-        while (Dep.size() != Level) {
+        while (Dep.size() < Level) {
           Dep.push_back('I');
         }
 
+        // Dependence analysis reports levels for the full enclosing loop nest.
+        // Keep only the suffix that corresponds to the selected perfect
+        // subnest.
+        if (Dep.size() > Level)
+          Dep.erase(Dep.begin(), Dep.end() - Level);
+
         // If all the elements of any direction vector have only '*', legality
         // can't be proven. Exit early to save compile time.
         if (all_of(Dep, equal_to('*'))) {
@@ -683,30 +688,31 @@ struct LoopInterchange {
     return processLoopList(LoopList);
   }
 
-  static SmallVector<SmallVector<Loop *, 8>, 4> collectRootToLeafPaths(Loop *Root) {
-    SmallVector<SmallVector<Loop *, 8>, 4> AllPaths;
-    // Stack stores {Loop, CurrentPath} pairs
-    SmallVector<std::pair<Loop *, SmallVector<Loop *, 8>>, 8> Stack;
-    Stack.push_back({Root, {Root}});
-    while (!Stack.empty()) {
-      auto [L, CurrentPath] = Stack.pop_back_val();
-      const auto &SubLoops = L->getSubLoops();
-      if (SubLoops.empty()) {
-        // Leaf node: save the path
-        AllPaths.push_back(CurrentPath);
-      } else {
-        for (Loop *Child : SubLoops) {
-          SmallVector<Loop *, 8> NewPath(CurrentPath);
-          NewPath.push_back(Child);
-          Stack.push_back({Child, NewPath});
-        }
+  static SmallVector<SmallVector<Loop *, 8>, 4>
+  collectPerfectNests(LoopNest &LN) {
+    SmallVector<SmallVector<Loop *, 8>, 4> LoopLists;
+    for (Loop *L : LN.getLoops()) {
+      if (!L->isInnermost())
+        continue;
+
+      SmallVector<Loop *, 8> LoopList;
+      Loop *Current = L;
+      while (true) {
+        LoopList.push_back(Current);
+        Loop *Parent = Current->getParentLoop();
+        if (!Parent || Parent->getSubLoops().size() != 1)
+          break;
+        Current = Parent;
       }
+      std::reverse(LoopList.begin(), LoopList.end());
+      if (LoopList.size() >= 2)
+        LoopLists.push_back(std::move(LoopList));
     }
-    return AllPaths;
+    return LoopLists;
   }
 
   bool run(LoopNest &LN) {
-    auto Paths = collectRootToLeafPaths(&LN.getOutermostLoop());
+    SmallVector<SmallVector<Loop *, 8>, 4> LoopLists = collectPerfectNests(LN);
     // Consider below kernel
     // for(int i=0; i<n; i++){  // Loop 1
     //     for(int j=0; j<m; j++){  // Loop 2
@@ -720,20 +726,20 @@ struct LoopInterchange {
     //         }
     //     }
     // }
     // Then Paths will contain:
     // - [Loop1, Loop2, Loop3]
     // - [Loop1, Loop4, Loop5]
     bool Changed = false;
-    for (auto &Path : Paths) {
+    for (SmallVector<Loop *, 8> &LoopList : LoopLists) {
       // Ensure minimum depth of the loop nest to do the interchange.
-      if (!hasSupportedLoopDepth(Path, *ORE))
+      if (!hasSupportedLoopDepth(LoopList, *ORE))
         continue;
       // Ensure computable loop nest.
-      if (!isComputableLoopNest(&AR->SE, Path)) {
+      if (!isComputableLoopNest(&AR->SE, LoopList)) {
         LLVM_DEBUG(dbgs() << "Not valid loop candidate for interchange\n");
         continue;
       }
-      Changed |= processLoopList(Path);
+      Changed |= processLoopList(LoopList);
     }
     return Changed;
   }
@@ -1511,11 +1517,11 @@ static bool areOuterLoopExitPHIsSupported(Loop *OuterLoop, Loop *InnerLoop) {
         continue;
 
       // The incoming value is defined in the outer loop latch. Currently we
-      // only support that in case the outer loop latch has a single predecessor.
-      // This guarantees that the outer loop latch is executed if and only if
-      // the inner loop is executed (because tightlyNested() guarantees that the
-      // outer loop header only branches to the inner loop or the outer loop
-      // latch).
+      // only support that in case the outer loop latch has a single
+      // predecessor. This guarantees that the outer loop latch is executed if
+      // and only if the inner loop is executed (because tightlyNested()
+      // guarantees that the outer loop header only branches to the inner loop
+      // or the outer loop latch).
       // FIXME: We could weaken this logic and allow multiple predecessors,
       //        if the values are produced outside the loop latch. We would need
       //        additional logic to update the PHI nodes in the exit block as

>From 002caa6bfbc3573d6fce8949a91489ae710a8ed9 Mon Sep 17 00:00:00 2001
From: rohgarg <rohgarg at qti.qualcomm.com>
Date: Wed, 10 Jun 2026 02:49:26 -0700
Subject: [PATCH 3/9] [LoopInterchange] Updated the tests for
 imperfect-loop-nests

---
 .../LoopInterchange/bail-out-one-loop.ll      |   1 +
 .../LoopInterchange/imperfect-loop-nest.ll    | 176 ------------------
 .../LoopInterchange/partially-perfect-loop.ll |  52 ++++++
@@ -720,20 +726,20 @@ struct LoopInterchange {
     //     }
     // }
-    // Then Paths will contain:
-    // - [Loop1, Loop2, Loop3]
-    // - [Loop1, Loop4, Loop5]
+    // Then LoopLists will contain:
+    // - [Loop2, Loop3]
+    // - [Loop4, Loop5]
     bool Changed = false;
 3 files changed, 53 insertions(+), 176 deletions(-)
 delete mode 100644 llvm/test/Transforms/LoopInterchange/imperfect-loop-nest.ll
